"""
***************************************************************************
    i_evapo_mh.py
    -------------
    Date                 : February 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "March 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

from qgis.PyQt.QtCore import QCoreApplication


def checkParameterValuesBeforeExecuting(alg, parameters, context):
    if alg.parameterAsBoolean(parameters, "-h", context) and alg.parameterAsLayer(
        parameters, "precipitation", context
    ):
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            "You can't use original Hargreaves flag and precipitation parameter together!",
        )
    if not alg.parameterAsBoolean(
        parameters, "-h", context
    ) and not alg.parameterAsLayer(parameters, "precipitation", context):
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            "If you don't use original Hargreaves flag, you must set the precipitation raster parameter!",
        )
    return True, None



"""
***************************************************************************
    __init__.py
    ---------------------
    Date                 : August 2013
    Copyright            : (C) 2013 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2013"
__copyright__ = "(C) 2013, Victor Olaya"

import logging

from qgis.PyQt import uic

uic.properties.logger.setLevel(logging.WARNING)
uic.uiparser.logger.setLevel(logging.WARNING)
uic.Compiler.qobjectcreator.logger.setLevel(logging.WARNING)



"""
***************************************************************************
    QgisAlgorithmProvider.py
    ---------------------
    Date                 : December 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "December 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os

from qgis.core import QgsApplication, QgsProcessingProvider, QgsRuntimeProfiler
from qgis.PyQt.QtCore import QCoreApplication

from .BarPlot import BarPlot
from .BoxPlot import BoxPlot
from .EliminateSelection import EliminateSelection
from .ExecuteSQL import ExecuteSQL
from .FieldPyculator import FieldsPyculator
from .Heatmap import Heatmap
from .HubDistanceLines import HubDistanceLines
from .HubDistancePoints import HubDistancePoints
from .HypsometricCurves import HypsometricCurves
from .IdwInterpolation import IdwInterpolation
from .KNearestConcaveHull import KNearestConcaveHull
from .LinesToPolygons import LinesToPolygons
from .MeanAndStdDevPlot import MeanAndStdDevPlot
from .PointDistance import PointDistance
from .PointsDisplacement import PointsDisplacement
from .PointsFromLines import PointsFromLines
from .PolarPlot import PolarPlot
from .RandomExtractWithinSubsets import RandomExtractWithinSubsets
from .RandomPointsAlongLines import RandomPointsAlongLines
from .RandomPointsLayer import RandomPointsLayer
from .RandomPointsPolygons import RandomPointsPolygons
from .RandomSelectionWithinSubsets import RandomSelectionWithinSubsets
from .RasterCalculator import RasterCalculator
from .RasterLayerHistogram import RasterLayerHistogram
from .RectanglesOvalsDiamondsVariable import RectanglesOvalsDiamondsVariable
from .RegularPoints import RegularPoints
from .Relief import Relief
from .SelectByAttribute import SelectByAttribute
from .SelectByExpression import SelectByExpression
from .SetRasterStyle import SetRasterStyle
from .SetVectorStyle import SetVectorStyle
from .StatisticsByCategories import StatisticsByCategories
from .TextToFloat import TextToFloat
from .TinInterpolation import TinInterpolation
from .TopoColors import TopoColor
from .VariableDistanceBuffer import VariableDistanceBuffer
from .VectorLayerHistogram import VectorLayerHistogram
from .VectorLayerScatterplot import VectorLayerScatterplot
from .VectorLayerScatterplot3D import VectorLayerScatterplot3D


class QgisAlgorithmProvider(QgsProcessingProvider):
    fieldMappingParameterName = QCoreApplication.translate(
        "Processing", "Fields Mapper"
    )

    def __init__(self):
        super().__init__()
        QgsApplication.processingRegistry().addAlgorithmAlias(
            "qgis:rectanglesovalsdiamondsfixed", "native:rectanglesovalsdiamonds"
        )

    def getAlgs(self):
        algs = [
            BarPlot(),
            BoxPlot(),
            EliminateSelection(),
            ExecuteSQL(),
            FieldsPyculator(),
            Heatmap(),
            HubDistanceLines(),
            HubDistancePoints(),
            HypsometricCurves(),
            IdwInterpolation(),
            KNearestConcaveHull(),
            LinesToPolygons(),
            MeanAndStdDevPlot(),
            PointDistance(),
            PointsDisplacement(),
            PointsFromLines(),
            PolarPlot(),
            RandomExtractWithinSubsets(),
            RandomPointsAlongLines(),
            RandomPointsLayer(),
            RandomPointsPolygons(),
            RandomSelectionWithinSubsets(),
            RasterCalculator(),
            RasterLayerHistogram(),
            RectanglesOvalsDiamondsVariable(),
            RegularPoints(),
            Relief(),
            SelectByAttribute(),
            SelectByExpression(),
            SetRasterStyle(),
            SetVectorStyle(),
            StatisticsByCategories(),
            TextToFloat(),
            TinInterpolation(),
            TopoColor(),
            VariableDistanceBuffer(),
            VectorLayerHistogram(),
            VectorLayerScatterplot(),
            VectorLayerScatterplot3D(),
        ]

        return algs

    def id(self):
        return "qgis"

    def helpId(self):
        return "qgis"

    def name(self):
        return "QGIS"

    def icon(self):
        return QgsApplication.getThemeIcon("/providerQgis.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("providerQgis.svg")

    def loadAlgorithms(self):
        for a in self.getAlgs():
            self.addAlgorithm(a)

    def load(self):
        with QgsRuntimeProfiler.profile("QGIS Python Provider"):
            success = super().load()

        return success

    def unload(self):
        super().unload()

    def supportsNonFileBasedOutput(self):
        return True

    def flags(self):
        return QgsProcessingProvider.Flag.FlagCompatibleWithVirtualRaster



"""
***************************************************************************
    grass_algorithms_imagery_test.py
    ------------------------------
    Date                 : May 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : mederic dot ribreux at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "May 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

import shutil

import AlgorithmsTestBase
import nose2
from grassprovider.grass_provider import GrassProvider
from grassprovider.grass_utils import GrassUtils
from qgis.core import QgsApplication
from qgis.testing import QgisTestCase, start_app


class TestGrassAlgorithmsImageryTest(QgisTestCase, AlgorithmsTestBase.AlgorithmsTest):
    @classmethod
    def setUpClass(cls):
        start_app()
        cls.provider = GrassProvider()
        QgsApplication.processingRegistry().addProvider(cls.provider)
        cls.cleanup_paths = []

        assert GrassUtils.installedVersion()

    @classmethod
    def tearDownClass(cls):
        QgsApplication.processingRegistry().removeProvider(cls.provider)
        for path in cls.cleanup_paths:
            shutil.rmtree(path)

    def definition_file(self):
        return "grass_algorithms_imagery_tests.yaml"


if __name__ == "__main__":
    nose2.main()



"""
***************************************************************************
    RegularPoints.py
    ---------------------
    Date                 : September 2014
    Copyright            : (C) 2014 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "September 2014"
__copyright__ = "(C) 2014, Alexander Bruy"

import os
from math import sqrt
from random import seed, uniform

from qgis.core import (
    QgsApplication,
    QgsFeature,
    QgsFeatureSink,
    QgsField,
    QgsFields,
    QgsGeometry,
    QgsPoint,
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterCrs,
    QgsProcessingParameterDistance,
    QgsProcessingParameterExtent,
    QgsProcessingParameterFeatureSink,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QMetaType
from qgis.PyQt.QtGui import QIcon

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class RegularPoints(QgisAlgorithm):
    EXTENT = "EXTENT"
    SPACING = "SPACING"
    INSET = "INSET"
    RANDOMIZE = "RANDOMIZE"
    IS_SPACING = "IS_SPACING"
    OUTPUT = "OUTPUT"
    CRS = "CRS"

    def icon(self):
        return QgsApplication.getThemeIcon("/algorithms/mAlgorithmRegularPoints.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("/algorithms/mAlgorithmRegularPoints.svg")

    def group(self):
        return self.tr("Vector creation")

    def groupId(self):
        return "vectorcreation"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterExtent(
                self.EXTENT, self.tr("Input extent"), optional=False
            )
        )
        self.addParameter(
            QgsProcessingParameterDistance(
                self.SPACING,
                self.tr("Point spacing/count"),
                100,
                self.CRS,
                False,
                0.000001,
            )
        )
        self.addParameter(
            QgsProcessingParameterDistance(
                self.INSET,
                self.tr("Initial inset from corner (LH side)"),
                0.0,
                self.CRS,
                False,
                0.0,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.RANDOMIZE, self.tr("Apply random offset to point spacing"), False
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.IS_SPACING, self.tr("Use point spacing"), True
            )
        )
        self.addParameter(
            QgsProcessingParameterCrs(
                self.CRS, self.tr("Output layer CRS"), "ProjectCrs"
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr("Regular points"),
                QgsProcessing.SourceType.TypeVectorPoint,
            )
        )

    def name(self):
        return "regularpoints"

    def displayName(self):
        return self.tr("Regular points")

    def shortDescription(self):
        return self.tr(
            "Creates a point layer with a given number of regular points, all of them within a given extent."
        )

    def shortHelpString(self):
        return self.tr(
            "This algorithm creates a point layer with a given number of regular points, all of them within a given extent."
        )

    def processAlgorithm(self, parameters, context, feedback):
        spacing = self.parameterAsDouble(parameters, self.SPACING, context)
        inset = self.parameterAsDouble(parameters, self.INSET, context)
        randomize = self.parameterAsBoolean(parameters, self.RANDOMIZE, context)
        isSpacing = self.parameterAsBoolean(parameters, self.IS_SPACING, context)
        crs = self.parameterAsCrs(parameters, self.CRS, context)
        extent = self.parameterAsExtent(parameters, self.EXTENT, context, crs)

        fields = QgsFields()
        fields.append(QgsField("id", QMetaType.Type.Int, "", 10, 0))

        (sink, dest_id) = self.parameterAsSink(
            parameters, self.OUTPUT, context, fields, QgsWkbTypes.Type.Point, crs
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        if randomize:
            seed()

        area = extent.width() * extent.height()
        if isSpacing:
            pSpacing = spacing
        else:
            pSpacing = sqrt(area / spacing)

        f = QgsFeature()
        f.initAttributes(1)
        f.setFields(fields)

        count = 0
        id = 0
        total = 100.0 / (area / pSpacing)
        y = extent.yMaximum() - inset

        extent_geom = QgsGeometry.fromRect(extent)
        extent_engine = QgsGeometry.createGeometryEngine(extent_geom.constGet())
        extent_engine.prepareGeometry()

        while y >= extent.yMinimum():
            x = extent.xMinimum() + inset
            while x <= extent.xMaximum():
                if feedback.isCanceled():
                    break

                if randomize:
                    geom = QgsGeometry(
                        QgsPoint(
                            uniform(x - (pSpacing / 2.0), x + (pSpacing / 2.0)),
                            uniform(y - (pSpacing / 2.0), y + (pSpacing / 2.0)),
                        )
                    )
                else:
                    geom = QgsGeometry(QgsPoint(x, y))

                if extent_engine.intersects(geom.constGet()):
                    f.setAttributes([id])
                    f.setGeometry(geom)
                    sink.addFeature(f, QgsFeatureSink.Flag.FastInsert)
                    x += pSpacing
                    id += 1

                count += 1
                feedback.setProgress(int(count * total))

            y = y - pSpacing

        sink.finalize()
        return {self.OUTPUT: dest_id}



"""
***************************************************************************
    EquivalentNumField.py
    ---------------------
    Date                 : January 2013
    Copyright            : (C) 2013 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "January 2013"
__copyright__ = "(C) 2013, Victor Olaya"

import warnings

from qgis.core import (
    QgsProcessingException,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterString,
)
from qgis.PyQt.QtCore import QCoreApplication

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm
from processing.tools import vector


class VectorLayerScatterplot3D(QgisAlgorithm):
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    XFIELD = "XFIELD"
    YFIELD = "YFIELD"
    ZFIELD = "ZFIELD"
    TITLE = "TITLE"
    XAXIS_TITLE = "XAXIS_TITLE"
    YAXIS_TITLE = "YAXIS_TITLE"
    ZAXIS_TITLE = "ZAXIS_TITLE"

    def group(self):
        return self.tr("Plots")

    def groupId(self):
        return "plots"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFeatureSource(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.XFIELD,
                self.tr("X attribute"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Numeric,
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.YFIELD,
                self.tr("Y attribute"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Numeric,
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.ZFIELD,
                self.tr("Z attribute"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Numeric,
            )
        )

        self.addParameter(
            QgsProcessingParameterString(self.TITLE, self.tr("Title"), optional=True)
        )

        self.addParameter(
            QgsProcessingParameterString(
                self.XAXIS_TITLE, self.tr("X-axis title"), optional=True
            )
        )

        self.addParameter(
            QgsProcessingParameterString(
                self.YAXIS_TITLE, self.tr("Y-axis title"), optional=True
            )
        )

        self.addParameter(
            QgsProcessingParameterString(
                self.ZAXIS_TITLE, self.tr("Z-axis title"), optional=True
            )
        )

        self.addParameter(
            QgsProcessingParameterFileDestination(
                self.OUTPUT, self.tr("Scatterplot 3D"), self.tr("HTML files (*.html)")
            )
        )

    def name(self):
        return "scatter3dplot"

    def displayName(self):
        return self.tr("Vector layer scatterplot 3D")

    def shortDescription(self):
        return self.tr("Creates a 3D scatter plot for a vector layer.")

    def shortHelpString(self):
        return self.tr("This algorithm creates a 3D scatter plot for a vector layer.")

    def processAlgorithm(self, parameters, context, feedback):
        try:
            # importing plotly throws Python warnings from within the library - filter these out
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=ResourceWarning)
                warnings.filterwarnings("ignore", category=ImportWarning)
                import plotly as plt
                import plotly.graph_objs as go
        except ImportError:
            raise QgsProcessingException(
                QCoreApplication.translate(
                    "VectorLayerScatterplot3D",
                    "This algorithm requires the Python “plotly” library. Please install this library and try again.",
                )
            )

        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        xfieldname = self.parameterAsString(parameters, self.XFIELD, context)
        yfieldname = self.parameterAsString(parameters, self.YFIELD, context)
        zfieldname = self.parameterAsString(parameters, self.ZFIELD, context)

        title = self.parameterAsString(parameters, self.TITLE, context)
        xaxis_title = self.parameterAsString(parameters, self.XAXIS_TITLE, context)
        yaxis_title = self.parameterAsString(parameters, self.YAXIS_TITLE, context)
        zaxis_title = self.parameterAsString(parameters, self.ZAXIS_TITLE, context)

        if title.strip() == "":
            title = None
        # Passing None for axis titles to Plotly does not work in this case: Unlike in 2D charts, Plotly would use x, y, z as axis titles.
        # If our users enter space, we still get invisible axes titles, resulting in the same behavior as for the other algs.
        if xaxis_title == "":
            xaxis_title = xfieldname
        if yaxis_title == "":
            yaxis_title = yfieldname
        if zaxis_title == "":
            zaxis_title = zfieldname

        output = self.parameterAsFileOutput(parameters, self.OUTPUT, context)

        values = vector.values(source, xfieldname, yfieldname, zfieldname)

        data = [
            go.Scatter3d(
                x=values[xfieldname],
                y=values[yfieldname],
                z=values[zfieldname],
                mode="markers",
            )
        ]

        fig = go.Figure(
            data=data,
            layout_title_text=title,
            layout_scene_xaxis_title=xaxis_title,
            layout_scene_yaxis_title=yaxis_title,
            layout_scene_zaxis_title=zaxis_title,
        )

        fig.write_html(output)

        return {self.OUTPUT: output}



# The PEP 484 type hints stub file for the _server module.
#
# Generated by SIP 6.15.3


import collections, re, typing, enum

try:
    from warnings import deprecated
except ImportError:
    pass

import PyQt6.sip

from PyQt6 import QtCore
from PyQt6 import QtGui
from PyQt6 import QtXml
from PyQt6 import QtNetwork
from PyQt6 import QtWidgets
from PyQt6 import QtSql
from PyQt6 import QtPrintSupport
from PyQt6 import QtPositioning
from PyQt6 import QtMultimedia
from PyQt6 import QtSerialPort
from qgis import _core

# Support for QDate, QDateTime and QTime.
import datetime

# Convenient type aliases.
PYQT_SIGNAL = typing.Union[QtCore.pyqtSignal, QtCore.pyqtBoundSignal]
PYQT_SLOT = typing.Union[collections.abc.Callable[..., Any], QtCore.pyqtBoundSignal]


class QgsCapabilitiesCache(QtCore.QObject):

    def __init__(self, size: int) -> None: ...

    def removeCapabilitiesDocument(self, path: str|None) -> None: ...
    def insertCapabilitiesDocument(self, configFilePath: str|None, key: str|None, doc: QtXml.QDomDocument|None) -> None: ...
    def searchCapabilitiesDocument(self, configFilePath: str|None, key: str|None) -> QtXml.QDomDocument|None: ...


class QgsConfigCache(QtCore.QObject):

    def __init__(self, settings: 'QgsServerSettings|None') -> None: ...

    def removeChangedEntries(self) -> None: ...
    def removeChangedEntry(self, path: str|None) -> None: ...
    projectRemovedFromCache: typing.ClassVar[QtCore.pyqtSignal]
    def projects(self) -> list[_core.QgsProject]: ...
    def strategyName(self) -> str: ...
    def project(self, path: str|None, settings: 'QgsServerSettings|None' = ...) -> _core.QgsProject|None: ...
    def removeEntry(self, path: str|None) -> None: ...
    @staticmethod
    def instance() -> 'QgsConfigCache|None': ...
    @staticmethod
    def initialize(settings: 'QgsServerSettings|None') -> None: ...


class QgsServerApi(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, serverIface: 'QgsServerInterface|None') -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerApi') -> None: ...

    def serverIface(self) -> 'QgsServerInterface|None': ...
    def executeRequest(self, context: 'QgsServerApiContext') -> None: ...
    def accept(self, url: QtCore.QUrl) -> bool: ...
    def rootPath(self) -> str: ...
    def version(self) -> str: ...
    def description(self) -> str: ...
    def name(self) -> str: ...


class QgsServerApiContext(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, apiRootPath: str|None, request: 'QgsServerRequest|None', response: 'QgsServerResponse|None', project: _core.QgsProject|None, serverInterface: 'QgsServerInterface|None') -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerApiContext') -> None: ...

    def handlerPath(self) -> str: ...
    def setRequest(self, request: 'QgsServerRequest|None') -> None: ...
    def apiRootPath(self) -> str: ...
    def matchedPath(self) -> str: ...
    def serverInterface(self) -> 'QgsServerInterface|None': ...
    def setProject(self, project: _core.QgsProject|None) -> None: ...
    def project(self) -> _core.QgsProject|None: ...
    def response(self) -> 'QgsServerResponse|None': ...
    def request(self) -> 'QgsServerRequest|None': ...


class QgsServerLogger(_core.QgsMessageLogConsole):

    def __init__(self) -> None: ...

    @typing.overload
    def logMessage(self, message: str|None, tag: str|None, level: _core.Qgis.MessageLevel) -> None: ...
    @typing.overload
    def logMessage(self, message: str|None, tag: str|None, level: _core.Qgis.MessageLevel, format: _core.Qgis.StringFormat) -> None: ...
    def setLogStderr(self) -> None: ...
    def setLogFile(self, filename: str|None = ...) -> None: ...
    def setLogLevel(self, level: _core.Qgis.MessageLevel) -> None: ...
    def logLevel(self) -> _core.Qgis.MessageLevel: ...
    @staticmethod
    def instance() -> 'QgsServerLogger|None': ...


class QgsServerOgcApi(QgsServerApi):

    class ContentType(enum.IntEnum):
        GEOJSON = ... # type: QgsServerOgcApi.ContentType
        OPENAPI3 = ... # type: QgsServerOgcApi.ContentType
        JSON = ... # type: QgsServerOgcApi.ContentType
        HTML = ... # type: QgsServerOgcApi.ContentType
        XML = ... # type: QgsServerOgcApi.ContentType

    class Rel(enum.IntEnum):
        alternate = ... # type: QgsServerOgcApi.Rel
        describedBy = ... # type: QgsServerOgcApi.Rel
        collection = ... # type: QgsServerOgcApi.Rel
        item = ... # type: QgsServerOgcApi.Rel
        self = ... # type: QgsServerOgcApi.Rel
        service_desc = ... # type: QgsServerOgcApi.Rel
        service_doc = ... # type: QgsServerOgcApi.Rel
        prev = ... # type: QgsServerOgcApi.Rel
        next = ... # type: QgsServerOgcApi.Rel
        license = ... # type: QgsServerOgcApi.Rel
        items = ... # type: QgsServerOgcApi.Rel
        conformance = ... # type: QgsServerOgcApi.Rel
        data = ... # type: QgsServerOgcApi.Rel

    staticMetaObject = ... # type: QtCore.QMetaObject

    @typing.overload
    def __init__(self, serverIface: 'QgsServerInterface|None', rootPath: str|None, name: str|None, description: str|None = ..., version: str|None = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerOgcApi') -> None: ...

    @staticmethod
    def mimeType(contentType: 'QgsServerOgcApi.ContentType') -> typing.Any: ...
    @staticmethod
    def contentTypeFromExtension(extension: typing.Any) -> 'QgsServerOgcApi.ContentType': ...
    @staticmethod
    @deprecated("""""")
    def contenTypeFromExtension(extension: typing.Any) -> 'QgsServerOgcApi.ContentType': ...
    @staticmethod
    def contentTypeToExtension(ct: 'QgsServerOgcApi.ContentType') -> str: ...
    @staticmethod
    def contentTypeToStdString(ct: 'QgsServerOgcApi.ContentType') -> typing.Any: ...
    @staticmethod
    def contentTypeToString(ct: 'QgsServerOgcApi.ContentType') -> str: ...
    @staticmethod
    def relToString(rel: 'QgsServerOgcApi.Rel') -> typing.Any: ...
    @staticmethod
    def sanitizeUrl(url: QtCore.QUrl) -> QtCore.QUrl: ...
    def registerHandler(self, handler: 'QgsServerOgcApiHandler|None') -> None: ...
    def executeRequest(self, context: QgsServerApiContext) -> None: ...
    def rootPath(self) -> str: ...
    def version(self) -> str: ...
    def description(self) -> str: ...
    def name(self) -> str: ...


class QgsServerOgcApiHandler(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerOgcApiHandler') -> None: ...

    def setContentTypes(self, contentTypes: collections.abc.Iterable[int]) -> None: ...
    @staticmethod
    def layerFromCollectionId(context: QgsServerApiContext, collectionId: str|None) -> _core.QgsVectorLayer|None: ...
    @staticmethod
    def parentLink(url: QtCore.QUrl, levels: int = ...) -> str: ...
    def contentTypeFromRequest(self, request: 'QgsServerRequest|None') -> QgsServerOgcApi.ContentType: ...
    def staticPath(self, context: QgsServerApiContext) -> str: ...
    def templatePath(self, context: QgsServerApiContext) -> str: ...
    def href(self, context: QgsServerApiContext, extraPath: str|None = ..., extension: str|None = ...) -> typing.Any: ...
    def write(self, data: typing.Any, context: QgsServerApiContext, htmlMetadata: dict[str|None, typing.Any] = ...) -> None: ...
    def contentTypeForAccept(self, accept: str|None) -> str: ...
    def values(self, context: QgsServerApiContext) -> dict[str, typing.Any]: ...
    def handleRequest(self, context: QgsServerApiContext) -> None: ...
    def defaultContentType(self) -> QgsServerOgcApi.ContentType: ...
    def tags(self) -> list[str]: ...
    def linkType(self) -> QgsServerOgcApi.Rel: ...
    def linkTitle(self) -> typing.Any: ...
    def description(self) -> typing.Any: ...
    def summary(self) -> typing.Any: ...
    def parameters(self, context: QgsServerApiContext) -> list['QgsServerQueryStringParameter']: ...
    def operationId(self) -> typing.Any: ...
    def path(self) -> QtCore.QRegularExpression: ...


class QgsServerStaticHandler(QgsServerOgcApiHandler):

    @typing.overload
    def __init__(self, pathRegExp: str|None = ..., staticPathSuffix: str|None = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerStaticHandler') -> None: ...

    def linkType(self) -> QgsServerOgcApi.Rel: ...
    def linkTitle(self) -> typing.Any: ...
    def description(self) -> typing.Any: ...
    def summary(self) -> typing.Any: ...
    def operationId(self) -> typing.Any: ...
    def path(self) -> QtCore.QRegularExpression: ...
    def handleRequest(self, context: QgsServerApiContext) -> None: ...


class QgsServerParameterDefinition(PyQt6.sip.wrapper):

    mDefaultValue = ... # type: typing.Any
    mType = ... # type: QtCore.QMetaType.Type
    mValue = ... # type: typing.Any

    @typing.overload
    def __init__(self, type: QtCore.QMetaType.Type = ..., defaultValue: typing.Any = ...) -> None: ...
    @typing.overload
    @deprecated("""""")
    def __init__(self, type: typing.Any, defaultValue: typing.Any = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerParameterDefinition') -> None: ...

    @staticmethod
    def raiseError(msg: str|None) -> None: ...
    def loadUrl(self) -> typing.Tuple[str, bool]: ...
    def toUrl(self) -> typing.Tuple[QtCore.QUrl, bool]: ...
    def toColor(self) -> typing.Tuple[QtGui.QColor, bool]: ...
    def toBool(self) -> bool: ...
    def toDouble(self) -> typing.Tuple[float, bool]: ...
    def toInt(self) -> typing.Tuple[int, bool]: ...
    def toRectangle(self) -> typing.Tuple[_core.QgsRectangle, bool]: ...
    def toExpressionList(self) -> list[str]: ...
    def toOgcFilterList(self) -> list[str]: ...
    def toGeomList(self, delimiter: str = ..., skipEmptyParts: bool = ...) -> typing.Tuple[list[_core.QgsGeometry], bool]: ...
    def toColorList(self, delimiter: str = ..., skipEmptyParts: bool = ...) -> typing.Tuple[list[QtGui.QColor], bool]: ...
    def toDoubleList(self, delimiter: str = ..., skipEmptyParts: bool = ...) -> typing.Tuple[list[float], bool]: ...
    def toIntList(self, delimiter: str = ..., skipEmptyParts: bool = ...) -> typing.Tuple[list[int], bool]: ...
    def toStringList(self, delimiter: str = ..., skipEmptyParts: bool = ...) -> list[str]: ...
    def toString(self, defaultValue: bool = ...) -> str: ...
    def isValid(self) -> bool: ...
    def typeName(self) -> str: ...


class QgsServerParameter(QgsServerParameterDefinition):

    class Name(enum.IntEnum):
        UNKNOWN = ... # type: QgsServerParameter.Name
        SERVICE = ... # type: QgsServerParameter.Name
        VERSION_SERVICE = ... # type: QgsServerParameter.Name
        REQUEST = ... # type: QgsServerParameter.Name
        MAP = ... # type: QgsServerParameter.Name
        FILE_NAME = ... # type: QgsServerParameter.Name

    mName = ... # type: 'QgsServerParameter.Name'
    staticMetaObject = ... # type: QtCore.QMetaObject

    @typing.overload
    def __init__(self, name: 'QgsServerParameter.Name' = ..., type: QtCore.QMetaType.Type = ..., defaultValue: typing.Any = ...) -> None: ...
    @typing.overload
    @deprecated("""""")
    def __init__(self, name: 'QgsServerParameter.Name', type: typing.Any, defaultValue: typing.Any = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerParameter') -> None: ...

    @typing.overload
    @staticmethod
    def name(name: 'QgsServerParameter.Name') -> str: ...
    @typing.overload
    @staticmethod
    def name(name: str|None) -> 'QgsServerParameter.Name': ...
    def raiseError(self) -> None: ...


class QgsServerParameters(PyQt6.sip.wrapper):

    staticMetaObject = ... # type: QtCore.QMetaObject

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, query: QtCore.QUrlQuery) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerParameters') -> None: ...

    def loadParameter(self, name: str|None, value: str|None) -> bool: ...
    def version(self) -> str: ...
    def fileName(self) -> str: ...
    def map(self) -> str: ...
    def request(self) -> str: ...
    def service(self) -> str: ...
    def toMap(self) -> dict[str, str]: ...
    def urlQuery(self) -> QtCore.QUrlQuery: ...
    def value(self, key: str|None) -> str: ...
    @typing.overload
    def remove(self, key: str|None) -> None: ...
    @typing.overload
    def remove(self, name: QgsServerParameter.Name) -> None: ...
    def add(self, key: str|None, value: str|None) -> None: ...
    def clear(self) -> None: ...
    def load(self, query: QtCore.QUrlQuery) -> None: ...


class QgsServerFeatureId(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerFeatureId') -> None: ...

    @staticmethod
    def pkSeparator() -> str: ...
    @staticmethod
    def getExpressionFromServerFid(serverFid: str|None, provider: _core.QgsVectorDataProvider|None) -> str: ...
    @staticmethod
    def updateFeatureRequestFromServerFids(featureRequest: _core.QgsFeatureRequest, serverFids: collections.abc.Iterable[str|None], provider: _core.QgsVectorDataProvider|None) -> _core.QgsFeatureRequest: ...
    @staticmethod
    def getServerFid(feature: _core.QgsFeature, pkAttributes: collections.abc.Iterable[int]) -> str: ...


class QgsServerQueryStringParameter(PyQt6.sip.wrapper):

    class Type(enum.Enum):
        String = ... # type: QgsServerQueryStringParameter.Type
        Integer = ... # type: QgsServerQueryStringParameter.Type
        Double = ... # type: QgsServerQueryStringParameter.Type
        Boolean = ... # type: QgsServerQueryStringParameter.Type
        List = ... # type: QgsServerQueryStringParameter.Type

    staticMetaObject = ... # type: QtCore.QMetaObject

    @typing.overload
    def __init__(self, name: str|None, required: bool = ..., type: 'QgsServerQueryStringParameter.Type' = ..., description: str|None = ..., defaultValue: typing.Any = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerQueryStringParameter') -> None: ...

    def setHidden(self, hidden: bool) -> None: ...
    def hidden(self) -> bool: ...
    def setDescription(self, description: str|None) -> None: ...
    def name(self) -> str: ...
    @staticmethod
    def typeName(type: 'QgsServerQueryStringParameter.Type') -> str: ...
    def description(self) -> str: ...
    def value(self, context: QgsServerApiContext) -> typing.Any: ...


class QgsServerSettingsEnv(QtCore.QObject):

    class EnvVar(enum.IntEnum):
        QGIS_OPTIONS_PATH = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_PARALLEL_RENDERING = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_MAX_THREADS = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_LOG_LEVEL = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_LOG_FILE = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_LOG_STDERR = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_PROJECT_FILE = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_IGNORE_BAD_LAYERS = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_IGNORE_RENDERING_ERRORS = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_RETRY_BAD_LAYERS = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_CACHE_DIRECTORY = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_CACHE_SIZE = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_SHOW_GROUP_SEPARATOR = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_OVERRIDE_SYSTEM_LOCALE = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_WMS_MAX_HEIGHT = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_WMS_MAX_WIDTH = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_API_RESOURCES_DIRECTORY = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_API_WFS3_MAX_LIMIT = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_API_WFS3_ROOT_PATH = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_TRUST_LAYER_METADATA = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_FORCE_READONLY_LAYERS = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_DISABLE_GETPRINT = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_LANDING_PAGE_PROJECTS_DIRECTORIES = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_LANDING_PAGE_PROJECTS_PG_CONNECTIONS = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_LOG_PROFILE = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_SERVICE_URL = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_WMS_SERVICE_URL = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_WFS_SERVICE_URL = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_WCS_SERVICE_URL = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_WMTS_SERVICE_URL = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_LANDING_PAGE_PREFIX = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_PROJECT_CACHE_CHECK_INTERVAL = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_PROJECT_CACHE_STRATEGY = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_PROJECT_CACHE_SIZE = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_ALLOWED_EXTRA_SQL_TOKENS = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_APPLICATION_NAME = ... # type: QgsServerSettingsEnv.EnvVar
        QGIS_SERVER_CAPABILITIES_CACHE_SIZE = ... # type: QgsServerSettingsEnv.EnvVar

    class Source(enum.IntEnum):
        DEFAULT_VALUE = ... # type: QgsServerSettingsEnv.Source
        ENVIRONMENT_VARIABLE = ... # type: QgsServerSettingsEnv.Source
        INI_FILE = ... # type: QgsServerSettingsEnv.Source

    def __init__(self) -> None: ...


class QgsServerSettings(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerSettings') -> None: ...

    def capabilitiesCacheSize(self) -> int: ...
    @staticmethod
    def name(env: QgsServerSettingsEnv.EnvVar) -> str: ...
    def applicationName(self) -> str: ...
    def allowedExtraSqlTokens(self) -> list[str]: ...
    def projectCacheSize(self) -> int: ...
    def projectCacheStrategy(self) -> str: ...
    def projectCacheCheckInterval(self) -> int: ...
    def serviceUrl(self, service: str|None) -> str: ...
    def getPrintDisabled(self) -> bool: ...
    def forceReadOnlyLayers(self) -> bool: ...
    def trustLayerMetadata(self) -> bool: ...
    def retryBadLayers(self) -> bool: ...
    def ignoreRenderingErrors(self) -> bool: ...
    def ignoreBadLayers(self) -> bool: ...
    def apiWfs3RootPath(self) -> str: ...
    def apiWfs3MaxLimit(self) -> int: ...
    def apiResourcesDirectory(self) -> str: ...
    def landingPageBaseUrlPrefix(self) -> str: ...
    def landingPageProjectsPgConnections(self) -> str: ...
    def landingPageProjectsDirectories(self) -> str: ...
    def wmsMaxWidth(self) -> int: ...
    def wmsMaxHeight(self) -> int: ...
    def showGroupSeparator(self) -> bool: ...
    def overrideSystemLocale(self) -> str: ...
    def cacheDirectory(self) -> str: ...
    def cacheSize(self) -> int: ...
    def logStderr(self) -> bool: ...
    def logFile(self) -> str: ...
    def projectFile(self) -> str: ...
    def logProfile(self) -> bool: ...
    def logLevel(self) -> _core.Qgis.MessageLevel: ...
    def maxThreads(self) -> int: ...
    def parallelRendering(self) -> bool: ...
    def iniFile(self) -> str: ...
    def logSummary(self) -> None: ...
    @typing.overload
    def load(self) -> None: ...
    @typing.overload
    def load(self, envVarName: str|None) -> bool: ...


class QgsServiceModule(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServiceModule') -> None: ...

    def registerSelf(self, registry: 'QgsServiceRegistry', serverIface: 'QgsServerInterface|None' = ...) -> None: ...


class QgsServerRequest(PyQt6.sip.wrapper):

    class RequestHeader(enum.IntEnum):
        HOST = ... # type: QgsServerRequest.RequestHeader
        FORWARDED = ... # type: QgsServerRequest.RequestHeader
        X_FORWARDED_FOR = ... # type: QgsServerRequest.RequestHeader
        X_FORWARDED_HOST = ... # type: QgsServerRequest.RequestHeader
        X_FORWARDED_PROTO = ... # type: QgsServerRequest.RequestHeader
        X_QGIS_SERVICE_URL = ... # type: QgsServerRequest.RequestHeader
        X_QGIS_WMS_SERVICE_URL = ... # type: QgsServerRequest.RequestHeader
        X_QGIS_WFS_SERVICE_URL = ... # type: QgsServerRequest.RequestHeader
        X_QGIS_WCS_SERVICE_URL = ... # type: QgsServerRequest.RequestHeader
        X_QGIS_WMTS_SERVICE_URL = ... # type: QgsServerRequest.RequestHeader
        ACCEPT = ... # type: QgsServerRequest.RequestHeader
        USER_AGENT = ... # type: QgsServerRequest.RequestHeader
        AUTHORIZATION = ... # type: QgsServerRequest.RequestHeader

    class Method(enum.IntEnum):
        HeadMethod = ... # type: QgsServerRequest.Method
        PutMethod = ... # type: QgsServerRequest.Method
        GetMethod = ... # type: QgsServerRequest.Method
        PostMethod = ... # type: QgsServerRequest.Method
        DeleteMethod = ... # type: QgsServerRequest.Method
        PatchMethod = ... # type: QgsServerRequest.Method
        OptionsMethod = ... # type: QgsServerRequest.Method

    staticMetaObject = ... # type: QtCore.QMetaObject

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, url: str|None, method: 'QgsServerRequest.Method' = ..., headers: dict[str|None, str|None] = ...) -> None: ...
    @typing.overload
    def __init__(self, url: QtCore.QUrl, method: 'QgsServerRequest.Method' = ..., headers: dict[str|None, str|None] = ...) -> None: ...
    @typing.overload
    def __init__(self, other: 'QgsServerRequest') -> None: ...

    def setBaseUrl(self, url: QtCore.QUrl) -> None: ...
    def setOriginalUrl(self, url: QtCore.QUrl) -> None: ...
    def queryParameter(self, name: str|None, defaultValue: str|None = ...) -> str: ...
    def setMethod(self, method: 'QgsServerRequest.Method') -> None: ...
    def baseUrl(self) -> QtCore.QUrl: ...
    def originalUrl(self) -> QtCore.QUrl: ...
    def setUrl(self, url: QtCore.QUrl) -> None: ...
    def data(self) -> QtCore.QByteArray: ...
    def removeHeader(self, name: str|None) -> None: ...
    def headers(self) -> dict[str, str]: ...
    def setHeader(self, name: str|None, value: str|None) -> None: ...
    @typing.overload
    def header(self, name: str|None) -> str: ...
    @typing.overload
    def header(self, headerEnum: 'QgsServerRequest.RequestHeader') -> str: ...
    def removeParameter(self, key: str|None) -> None: ...
    def parameter(self, key: str|None, defaultValue: str|None = ...) -> str: ...
    def setParameter(self, key: str|None, value: str|None) -> None: ...
    def serverParameters(self) -> QgsServerParameters: ...
    def parameters(self) -> dict[str, str]: ...
    def method(self) -> 'QgsServerRequest.Method': ...
    def url(self) -> QtCore.QUrl: ...
    @staticmethod
    def methodToString(method: 'QgsServerRequest.Method') -> str: ...


class QgsBufferServerRequest(QgsServerRequest):

    @typing.overload
    def __init__(self, url: str|None, method: QgsServerRequest.Method = ..., headers: dict[str|None, str|None] = ..., data: QtCore.QByteArray|bytes|bytearray|memoryview = ...) -> None: ...
    @typing.overload
    def __init__(self, url: QtCore.QUrl, method: QgsServerRequest.Method = ..., headers: dict[str|None, str|None] = ..., data: QtCore.QByteArray|bytes|bytearray|memoryview = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsBufferServerRequest') -> None: ...

    def data(self) -> QtCore.QByteArray: ...


class QgsServerResponse(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerResponse') -> None: ...

    def feedback(self) -> _core.QgsFeedback|None: ...
    def truncate(self) -> None: ...
    def data(self) -> QtCore.QByteArray: ...
    def clear(self) -> None: ...
    def flush(self) -> None: ...
    def finish(self) -> None: ...
    def io(self) -> QtCore.QIODevice|None: ...
    @typing.overload
    def write(self, data: str|None) -> None: ...
    @typing.overload
    def write(self, byteArray: QtCore.QByteArray|bytes|bytearray|memoryview) -> int: ...
    @typing.overload
    def write(self, ex: 'QgsServerException') -> None: ...
    def sendError(self, code: int, message: str|None) -> None: ...
    def statusCode(self) -> int: ...
    def setStatusCode(self, code: int) -> None: ...
    def headersSent(self) -> bool: ...
    def headers(self) -> dict[str, str]: ...
    def header(self, key: str|None) -> str: ...
    def removeHeader(self, key: str|None) -> None: ...
    def setHeader(self, key: str|None, value: str|None) -> None: ...


class QgsBufferServerResponse(QgsServerResponse):

    def __init__(self) -> None: ...

    def body(self) -> QtCore.QByteArray: ...
    def truncate(self) -> None: ...
    def data(self) -> QtCore.QByteArray: ...
    def clear(self) -> None: ...
    def flush(self) -> None: ...
    def finish(self) -> None: ...
    def io(self) -> QtCore.QIODevice|None: ...
    def sendError(self, code: int, message: str|None) -> None: ...
    def statusCode(self) -> int: ...
    def setStatusCode(self, code: int) -> None: ...
    def headersSent(self) -> bool: ...
    def headers(self) -> dict[str, str]: ...
    def header(self, key: str|None) -> str: ...
    def removeHeader(self, key: str|None) -> None: ...
    def setHeader(self, key: str|None, value: str|None) -> None: ...


class QgsFcgiServerRequest(QgsServerRequest):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsFcgiServerRequest') -> None: ...

    def header(self, name: str|None) -> str: ...
    def hasError(self) -> bool: ...
    def data(self) -> QtCore.QByteArray: ...


class QgsRequestHandler(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, request: QgsServerRequest, response: QgsServerResponse) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsRequestHandler') -> None: ...

    def headersSent(self) -> bool: ...
    def format(self) -> str: ...
    def removeParameter(self, key: str|None) -> None: ...
    def parameter(self, key: str|None) -> str: ...
    def setParameter(self, key: str|None, value: str|None) -> None: ...
    def parameterMap(self) -> dict[str, str]: ...
    def statusCode(self) -> int: ...
    def setStatusCode(self, code: int) -> None: ...
    def path(self) -> str: ...
    def url(self) -> str: ...
    def data(self) -> QtCore.QByteArray: ...
    def body(self) -> QtCore.QByteArray: ...
    def clearBody(self) -> None: ...
    def exceptionRaised(self) -> bool: ...
    def appendBody(self, body: QtCore.QByteArray|bytes|bytearray|memoryview) -> None: ...
    def clear(self) -> None: ...
    def requestHeaders(self) -> dict[str, str]: ...
    def requestHeader(self, name: str|None) -> str: ...
    def removeRequestHeader(self, name: str|None) -> None: ...
    def setRequestHeader(self, name: str|None, value: str|None) -> None: ...
    def responseHeaders(self) -> dict[str, str]: ...
    def responseHeader(self, name: str|None) -> str: ...
    def removeResponseHeader(self, name: str|None) -> None: ...
    def setResponseHeader(self, name: str|None, value: str|None) -> None: ...
    def sendResponse(self) -> None: ...
    def setServiceException(self, ex: 'QgsServerException') -> None: ...


class QgsServer(PyQt6.sip.wrapper):

    def __init__(self) -> None: ...

    def initPython(self) -> None: ...
    def serverInterface(self) -> 'QgsServerInterface|None': ...
    def handleRequest(self, request: QgsServerRequest, response: QgsServerResponse, project: _core.QgsProject|None = ...) -> None: ...
    def putenv(self, var: str|None, val: str|None) -> None: ...


class QgsServerApiUtils(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerApiUtils') -> None: ...

    @staticmethod
    def appendMapParameter(path: str|None, requestUrl: QtCore.QUrl) -> str: ...
    @staticmethod
    @deprecated("""""")
    def crsToOgcUri(crs: _core.QgsCoordinateReferenceSystem) -> str: ...
    @staticmethod
    def publishedCrsList(project: _core.QgsProject|None) -> list[str]: ...
    @staticmethod
    def sanitizedFieldValue(value: str|None) -> str: ...
    @staticmethod
    def publishedWfsLayers(context: QgsServerApiContext) -> list[_core.QgsVectorLayer]: ...
    @staticmethod
    def parseCrs(bboxCrs: str|None) -> _core.QgsCoordinateReferenceSystem: ...
    @staticmethod
    def temporalExtent(layer: _core.QgsVectorLayer|None) -> list[typing.Any]: ...
    @staticmethod
    def temporalFilterExpression(layer: _core.QgsVectorLayer|None, interval: str|None) -> _core.QgsExpression: ...
    @staticmethod
    def fieldName(name: str|None, layer: _core.QgsVectorLayer|None) -> str: ...
    @staticmethod
    def parseTemporalDateTimeInterval(interval: str|None) -> _core.QgsDateTimeRange: ...
    @staticmethod
    def parseTemporalDateInterval(interval: str|None) -> _core.QgsDateRange: ...
    @staticmethod
    def temporalDimensions(layer: _core.QgsVectorLayer|None) -> list[_core.QgsServerWmsDimensionProperties.WmsDimensionInfo]: ...
    @staticmethod
    def parseBbox(bbox: str|None) -> _core.QgsRectangle: ...


class QgsServerException(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, message: str|None, responseCode: int = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerException') -> None: ...

    def formatResponse(self) -> typing.Tuple[QtCore.QByteArray, str]: ...
    def responseCode(self) -> int: ...


class QgsOgcServiceException(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, code: str|None, message: str|None, locator: str|None = ..., responseCode: int = ..., version: str|None = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsOgcServiceException') -> None: ...

    def formatResponse(self) -> typing.Tuple[QtCore.QByteArray, str]: ...
    def version(self) -> str: ...
    def locator(self) -> str: ...
    def code(self) -> str: ...
    def message(self) -> str: ...


class QgsServerFilter(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, serverInterface: 'QgsServerInterface|None') -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerFilter') -> None: ...

    def onSendResponse(self) -> bool: ...
    def onResponseComplete(self) -> bool: ...
    def onProjectReady(self) -> bool: ...
    def onRequestReady(self) -> bool: ...
    @deprecated("""""")
    def sendResponse(self) -> None: ...
    @deprecated("""""")
    def responseComplete(self) -> None: ...
    @deprecated("""""")
    def requestReady(self) -> None: ...
    def serverInterface(self) -> 'QgsServerInterface|None': ...


class QgsServerInterface(PyQt6.sip.wrapper):

    def __init__(self, a0: 'QgsServerInterface') -> None: ...

    def reloadSettings(self) -> None: ...
    def serviceRegistry(self) -> 'QgsServiceRegistry|None': ...
    def removeConfigCacheEntry(self, path: str|None) -> None: ...
    def setConfigFilePath(self, configFilePath: str|None) -> None: ...
    def configFilePath(self) -> str: ...
    def getEnv(self, name: str|None) -> str: ...
    def cacheManager(self) -> 'QgsServerCacheManager|None': ...
    def registerServerCache(self, serverCache: 'QgsServerCacheFilter|None', priority: int = ...) -> None: ...
    def accessControls(self) -> 'QgsAccessControl|None': ...
    def registerAccessControl(self, accessControl: 'QgsAccessControlFilter|None', priority: int = ...) -> None: ...
    def filters(self) -> typing.Any: ...
    def setFilters(self, filters: typing.Any) -> None: ...
    def registerFilter(self, filter: QgsServerFilter|None, priority: int = ...) -> None: ...
    def requestHandler(self) -> QgsRequestHandler|None: ...
    def capabilitiesCache(self) -> QgsCapabilitiesCache|None: ...


class QgsServerProjectUtils(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerProjectUtils') -> None: ...

    @staticmethod
    def serviceUrl(service: str|None, request: QgsServerRequest, settings: QgsServerSettings) -> str: ...
    @staticmethod
    def wmtsServiceUrl(project: _core.QgsProject, request: QgsServerRequest = ..., settings: QgsServerSettings = ...) -> str: ...
    @staticmethod
    def wcsLayerIds(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def wcsServiceUrl(project: _core.QgsProject, request: QgsServerRequest = ..., settings: QgsServerSettings = ...) -> str: ...
    @staticmethod
    def wfstDeleteLayerIds(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def wfstInsertLayerIds(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def wfstUpdateLayerIds(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def wfsLayerPrecision(project: _core.QgsProject, layerId: str|None) -> int: ...
    @staticmethod
    def wfsLayerIds(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def wfsServiceUrl(project: _core.QgsProject, request: QgsServerRequest = ..., settings: QgsServerSettings = ...) -> str: ...
    @staticmethod
    def wmsExtent(project: _core.QgsProject) -> _core.QgsRectangle: ...
    @staticmethod
    def wmsOutputCrsListAsOgcUrn(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def wmsOutputCrsList(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def wmsRestrictedLayers(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def wmsRootName(project: _core.QgsProject) -> str: ...
    @staticmethod
    def wmsServiceUrl(project: _core.QgsProject, request: QgsServerRequest = ..., settings: QgsServerSettings = ...) -> str: ...
    @staticmethod
    def wmsRestrictedComposers(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def wmsInspireMetadataDate(project: _core.QgsProject) -> str: ...
    @staticmethod
    def wmsInspireTemporalReference(project: _core.QgsProject) -> str: ...
    @staticmethod
    def wmsInspireMetadataUrlType(project: _core.QgsProject) -> str: ...
    @staticmethod
    def wmsInspireMetadataUrl(project: _core.QgsProject) -> str: ...
    @staticmethod
    def wmsInspireLanguage(project: _core.QgsProject) -> str: ...
    @staticmethod
    def wmsInspireActivate(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsFeatureInfoLayerAliasMap(project: _core.QgsProject) -> dict[str, str]: ...
    @staticmethod
    def wmsFeatureInfoSchema(project: _core.QgsProject) -> str: ...
    @staticmethod
    def wmsFeatureInfoDocumentElementNs(project: _core.QgsProject) -> str: ...
    @staticmethod
    def wmsFeatureInfoDocumentElement(project: _core.QgsProject) -> str: ...
    @staticmethod
    def wmsFeatureInfoPrecision(project: _core.QgsProject) -> int: ...
    @staticmethod
    def wmsSkipNameForGroup(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsAddLegendGroupsLegendGraphic(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsFeatureInfoSegmentizeWktGeometry(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsHTMLFeatureInfoUseOnlyMaptip(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsFeatureInfoUseAttributeFormSettings(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsFeatureInfoAddWktGeometry(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsInfoFormatSia2045(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsUseLayerIds(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsDefaultMapUnitsPerMm(project: _core.QgsProject) -> float: ...
    @staticmethod
    def wmsMaxAtlasFeatures(project: _core.QgsProject) -> int: ...
    @staticmethod
    def wmsRenderMapTiles(project: _core.QgsProject) -> bool: ...
    @staticmethod
    def wmsTileBuffer(project: _core.QgsProject) -> int: ...
    @staticmethod
    def wmsImageQuality(project: _core.QgsProject) -> int: ...
    @staticmethod
    def wmsMaxHeight(project: _core.QgsProject) -> int: ...
    @staticmethod
    def wmsMaxWidth(project: _core.QgsProject) -> int: ...
    @staticmethod
    def owsServiceAccessConstraints(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceFees(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceContactPhone(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceContactMail(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceContactPerson(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceContactPosition(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceContactOrganization(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceOnlineResource(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceKeywords(project: _core.QgsProject) -> list[str]: ...
    @staticmethod
    def owsServiceAbstract(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceTitle(project: _core.QgsProject) -> str: ...
    @staticmethod
    def owsServiceCapabilities(project: _core.QgsProject) -> bool: ...


class QgsService(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsService') -> None: ...

    def executeRequest(self, request: QgsServerRequest, response: QgsServerResponse, project: _core.QgsProject|None) -> None: ...
    def version(self) -> str: ...
    def name(self) -> str: ...


class QgsServiceRegistry(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServiceRegistry') -> None: ...

    def cleanUp(self) -> None: ...
    def init(self, nativeModulepath: str|None, serverIface: QgsServerInterface|None = ...) -> None: ...
    def unregisterService(self, name: str|None, version: str|None = ...) -> int: ...
    def getApi(self, name: str|None, version: str|None = ...) -> QgsServerApi|None: ...
    def unregisterApi(self, name: str|None, version: str|None = ...) -> int: ...
    def registerApi(self, api: QgsServerApi|None) -> bool: ...
    def registerService(self, service: QgsService|None) -> None: ...
    def getService(self, name: str|None, version: str|None = ...) -> QgsService|None: ...


class QgsFeatureFilterProviderGroup(_core.QgsFeatureFilterProvider):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsFeatureFilterProviderGroup') -> None: ...

    def addProvider(self, provider: _core.QgsFeatureFilterProvider|None) -> 'QgsFeatureFilterProviderGroup': ...
    def clone(self) -> 'QgsFeatureFilterProviderGroup|None': ...
    def layerAttributes(self, layer: _core.QgsVectorLayer|None, attributes: collections.abc.Iterable[str|None]) -> list[str]: ...
    def filterFeatures(self, layer: _core.QgsVectorLayer|None, filterFeatures: _core.QgsFeatureRequest) -> None: ...
    def isFilterThreadSafe(self) -> bool: ...


class QgsFeatureFilter(_core.QgsFeatureFilterProvider):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsFeatureFilter') -> None: ...

    def setFilter(self, layer: _core.QgsVectorLayer|None, expression: _core.QgsExpression) -> None: ...
    def clone(self) -> _core.QgsFeatureFilterProvider|None: ...
    def layerAttributes(self, layer: _core.QgsVectorLayer|None, attributes: collections.abc.Iterable[str|None]) -> list[str]: ...
    def filterFeatures(self, layer: _core.QgsVectorLayer|None, filterFeatures: _core.QgsFeatureRequest) -> None: ...
    def isFilterThreadSafe(self) -> bool: ...


class QgsStoreBadLayerInfo(_core.QgsProjectBadLayerHandler):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsStoreBadLayerInfo') -> None: ...

    def badLayerNames(self) -> dict[str, str]: ...
    def badLayers(self) -> list[str]: ...
    def handleBadLayers(self, layers: collections.abc.Iterable[QtXml.QDomNode]) -> None: ...


class QgsAccessControlFilter(PyQt6.sip.wrapper):

    class LayerPermissions(PyQt6.sip.wrapper):

        canDelete = ... # type: bool
        canInsert = ... # type: bool
        canRead = ... # type: bool
        canUpdate = ... # type: bool

        @typing.overload
        def __init__(self) -> None: ...
        @typing.overload
        def __init__(self, a0: 'QgsAccessControlFilter.LayerPermissions') -> None: ...

    @typing.overload
    def __init__(self, serverInterface: QgsServerInterface|None) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsAccessControlFilter') -> None: ...

    def cacheKey(self) -> str: ...
    def allowToEdit(self, layer: _core.QgsVectorLayer|None, feature: _core.QgsFeature) -> bool: ...
    def authorizedLayerAttributes(self, layer: _core.QgsVectorLayer|None, attributes: collections.abc.Iterable[str|None]) -> list[str]: ...
    def layerPermissions(self, layer: _core.QgsMapLayer|None) -> 'QgsAccessControlFilter.LayerPermissions': ...
    def layerFilterSubsetString(self, layer: _core.QgsVectorLayer|None) -> str: ...
    def layerFilterExpression(self, layer: _core.QgsVectorLayer|None) -> str: ...
    def serverInterface(self) -> QgsServerInterface|None: ...


class QgsAccessControl(_core.QgsFeatureFilterProvider):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, copy: 'QgsAccessControl') -> None: ...

    def registerAccessControl(self, accessControl: QgsAccessControlFilter|None, priority: int = ...) -> None: ...
    def fillCacheKey(self, cacheKey: collections.abc.Iterable[str|None]) -> bool: ...
    def allowToEdit(self, layer: _core.QgsVectorLayer|None, feature: _core.QgsFeature) -> bool: ...
    def layerDeletePermission(self, layer: _core.QgsVectorLayer|None) -> bool: ...
    def layerUpdatePermission(self, layer: _core.QgsVectorLayer|None) -> bool: ...
    def layerInsertPermission(self, layer: _core.QgsVectorLayer|None) -> bool: ...
    def layerReadPermission(self, layer: _core.QgsMapLayer|None) -> bool: ...
    def extraSubsetString(self, layer: _core.QgsVectorLayer|None) -> str: ...
    def unresolveFilterFeatures(self) -> None: ...
    def resolveFilterFeatures(self, layers: collections.abc.Iterable[_core.QgsMapLayer]) -> None: ...
    def clone(self) -> 'QgsAccessControl|None': ...
    def layerAttributes(self, layer: _core.QgsVectorLayer|None, attributes: collections.abc.Iterable[str|None]) -> list[str]: ...
    def filterFeatures(self, layer: _core.QgsVectorLayer|None, filterFeatures: _core.QgsFeatureRequest) -> None: ...
    def isFilterThreadSafe(self) -> bool: ...


class QgsServerCacheFilter(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, serverInterface: QgsServerInterface|None) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsServerCacheFilter') -> None: ...

    def deleteCachedImages(self, project: _core.QgsProject|None) -> bool: ...
    def deleteCachedImage(self, project: _core.QgsProject|None, request: QgsServerRequest, key: str|None) -> bool: ...
    def setCachedImage(self, img: QtCore.QByteArray|bytes|bytearray|memoryview, project: _core.QgsProject|None, request: QgsServerRequest, key: str|None) -> bool: ...
    def getCachedImage(self, project: _core.QgsProject|None, request: QgsServerRequest, key: str|None) -> QtCore.QByteArray: ...
    def deleteCachedDocuments(self, project: _core.QgsProject|None) -> bool: ...
    def deleteCachedDocument(self, project: _core.QgsProject|None, request: QgsServerRequest, key: str|None) -> bool: ...
    def setCachedDocument(self, doc: QtXml.QDomDocument|None, project: _core.QgsProject|None, request: QgsServerRequest, key: str|None) -> bool: ...
    def getCachedDocument(self, project: _core.QgsProject|None, request: QgsServerRequest, key: str|None) -> QtCore.QByteArray: ...


class QgsServerCacheManager(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, settings: QgsServerSettings = ...) -> None: ...
    @typing.overload
    def __init__(self, copy: 'QgsServerCacheManager') -> None: ...

    def registerServerCache(self, serverCache: QgsServerCacheFilter|None, priority: int = ...) -> None: ...
    def deleteCachedImages(self, project: _core.QgsProject|None) -> bool: ...
    def deleteCachedImage(self, project: _core.QgsProject|None, request: QgsServerRequest, accessControl: QgsAccessControl|None) -> bool: ...
    def setCachedImage(self, img: QtCore.QByteArray|bytes|bytearray|memoryview, project: _core.QgsProject|None, request: QgsServerRequest, accessControl: QgsAccessControl|None) -> bool: ...
    def getCachedImage(self, project: _core.QgsProject|None, request: QgsServerRequest, accessControl: QgsAccessControl|None) -> QtCore.QByteArray: ...
    def deleteCachedDocuments(self, project: _core.QgsProject|None) -> bool: ...
    def deleteCachedDocument(self, project: _core.QgsProject|None, request: QgsServerRequest, accessControl: QgsAccessControl|None) -> bool: ...
    def setCachedDocument(self, doc: QtXml.QDomDocument|None, project: _core.QgsProject|None, request: QgsServerRequest, accessControl: QgsAccessControl|None) -> bool: ...
    def getCachedDocument(self, doc: QtXml.QDomDocument|None, project: _core.QgsProject|None, request: QgsServerRequest, accessControl: QgsAccessControl|None) -> bool: ...



"""
***************************************************************************
    r_null.py
    ---------------------
    Date                 : February 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "February 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

from qgis.PyQt.QtCore import QCoreApplication


def checkParameterValuesBeforeExecuting(alg, parameters, context):
    """Verify if we have the right parameters"""
    if alg.parameterAsString(parameters, "setnull", context) or alg.parameterAsString(
        parameters, "null", context
    ):
        return True, None

    return False, QCoreApplication.translate(
        "GrassAlgorithmExt",
        "You need to set at least 'setnull' or 'null' parameters for this algorithm!",
    )


def processInputs(alg, parameters, context, feedback):
    """Prepare the GRASS import commands"""
    if "map" in alg.exportedLayers:
        return

    # We need to import without r.external
    alg.loadRasterLayerFromParameter("map", parameters, context, False)
    alg.postInputs(context)


def processCommand(alg, parameters, context, feedback):
    # We temporary remove the output 'sequence'
    alg.processCommand(parameters, context, feedback, True)


def processOutputs(alg, parameters, context, feedback):
    fileName = alg.parameterAsOutputLayer(parameters, "output", context)
    grassName = alg.exportedLayers["map"]
    alg.exportRasterLayer(grassName, fileName, False)



# /***************************************************************************
#  *                                                                         *
#  *   This program is free software; you can redistribute it and/or modify  *
#  *   it under the terms of the GNU General Public License as published by  *
#  *   the Free Software Foundation; either version 2 of the License, or     *
#  *   (at your option) any later version.                                   *
#  *                                                                         *
#  ***************************************************************************/

"""
GUI Utilities
"""

import os
from typing import Optional

from qgis.PyQt.QtGui import QIcon, QImage, QPixmap


class GuiUtils:
    """
    Utilities for GUI plugin components
    """

    ICON_CACHE: dict[str, QIcon] = {}

    @staticmethod
    def get_icon(icon: str) -> QIcon:
        """
        Returns a plugin icon
        :param icon: icon name (base part of file name)
        :return: QIcon
        """
        if icon in GuiUtils.ICON_CACHE:
            return GuiUtils.ICON_CACHE[icon]

        # prefer SVG files if present
        path = GuiUtils.get_icon_svg(icon)
        if path:
            res = QIcon(path)
            GuiUtils.ICON_CACHE[icon] = res
            return res

        pixmap = GuiUtils.get_icon_as_pixmap(icon)
        if pixmap is not None:
            res = QIcon(pixmap)
            GuiUtils.ICON_CACHE[icon] = res
            return res

        # return an invalid icon
        GuiUtils.ICON_CACHE[icon] = QIcon()
        return QIcon()

    @staticmethod
    def get_icon_svg(icon: str) -> str:
        """
        Returns a plugin icon's SVG file path
        :param icon: icon name (base part of file name)
        :return: icon svg path
        """
        path = os.path.join(os.path.dirname(__file__), "icons", icon + ".svg")
        if not os.path.exists(path):
            return ""

        return path

    @staticmethod
    def get_pixmap_path(icon: str) -> Optional[str]:
        """
        Returns the path to a pixmap icon
        """
        for suffix in (".png", ".gif", ".xpm"):
            path = os.path.join(os.path.dirname(__file__), "icons", icon + suffix)
            if os.path.exists(path):
                return path

        return None

    @staticmethod
    def get_icon_as_pixmap(icon: str) -> Optional[QPixmap]:
        """
        Returns a plugin icon's PNG file path
        :param icon: icon name (png file name)
        :return: icon png path
        """
        path = GuiUtils.get_pixmap_path(icon)
        if path is not None:
            im = QImage(path)
            return QPixmap.fromImage(im)

        return None

    @staticmethod
    def get_ui_file_path(file: str) -> str:
        """
        Returns a UI file's path
        :param file: file name (uifile name)
        :return: ui file path
        """
        path = os.path.join(os.path.dirname(__file__), "ui", file)
        if not os.path.exists(path):
            return ""

        return path



"""
***************************************************************************
    ogr2ogr.py
    ---------------------
    Date                 : November 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "November 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os

from qgis.core import (
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterString,
    QgsProcessingParameterVectorDestination,
)

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils


class ogr2ogr(GdalAlgorithm):
    INPUT = "INPUT"
    CONVERT_ALL_LAYERS = "CONVERT_ALL_LAYERS"
    OPTIONS = "OPTIONS"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT,
                self.tr("Input layer"),
                types=[QgsProcessing.SourceType.TypeVector],
            )
        )

        convert_all_layers_param = QgsProcessingParameterBoolean(
            self.CONVERT_ALL_LAYERS,
            self.tr("Convert all layers from dataset"),
            defaultValue=False,
        )
        convert_all_layers_param.setHelp(
            self.tr(
                "Use convert all layers to convert a whole dataset. "
                "Supported output formats for this option are GPKG and GML."
            )
        )
        self.addParameter(convert_all_layers_param)

        options_param = QgsProcessingParameterString(
            self.OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        options_param.setFlags(
            options_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(options_param)

        self.addParameter(
            QgsProcessingParameterVectorDestination(self.OUTPUT, self.tr("Converted"))
        )

    def name(self):
        return "convertformat"

    def displayName(self):
        return self.tr("Convert format")

    def group(self):
        return self.tr("Vector conversion")

    def groupId(self):
        return "vectorconversion"

    def commandName(self):
        return "ogr2ogr"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        input_details = self.getOgrCompatibleSource(
            self.INPUT, parameters, context, feedback, executing
        )
        convertAllLayers = self.parameterAsBoolean(
            parameters, self.CONVERT_ALL_LAYERS, context
        )

        if not convertAllLayers and not input_details.layer_name:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        options = self.parameterAsString(parameters, self.OPTIONS, context)
        outFile = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, outFile)

        output_details = GdalUtils.gdal_connection_details_from_uri(outFile, context)

        if output_details.format in ("SQLite", "GPKG") and os.path.isfile(
            output_details.connection_string
        ):
            raise QgsProcessingException(
                self.tr('Output file "{}" already exists.').format(
                    output_details.connection_string
                )
            )

        arguments = []
        if output_details.format:
            arguments.append(f"-f {output_details.format}")

        if input_details.open_options:
            arguments.extend(input_details.open_options_as_arguments())

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        if options:
            arguments.append(options)

        arguments.append(output_details.connection_string)
        arguments.append(input_details.connection_string)
        if not convertAllLayers:
            arguments.append(input_details.layer_name)

        return ["ogr2ogr", GdalUtils.escapeAndJoin(arguments)]

    def shortHelpString(self):
        return self.tr(
            "The algorithm converts simple features data between file formats.\n\n"
            "Use convert all layers to convert a whole dataset.\n"
            "Supported output formats for this option are:\n"
            "- GPKG\n"
            "- GML"
        )



"""
/***************************************************************************
Python Console for QGIS
                             -------------------
begin                : 2012-09-10
copyright            : (C) 2012 by Salvatore Larosa
email                : lrssvtml (at) gmail (dot) com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
Some portions of code were taken from https://code.google.com/p/pydee/
"""

from __future__ import annotations

import sys
from functools import partial
from typing import TYPE_CHECKING

from qgis.core import Qgis, QgsApplication, QgsSettings
from qgis.gui import QgsCodeEditorPython, QgsMessageBar
from qgis.PyQt import sip
from qgis.PyQt.Qsci import QsciScintilla
from qgis.PyQt.QtCore import (
    Q_ARG,
    QCoreApplication,
    QMetaObject,
    QObject,
    Qt,
    QThread,
    pyqtSlot,
)
from qgis.PyQt.QtGui import QColor, QKeySequence
from qgis.PyQt.QtWidgets import (
    QAction,
    QApplication,
    QGridLayout,
    QMenu,
    QShortcut,
    QSizePolicy,
    QSpacerItem,
)

if TYPE_CHECKING:
    from .console import PythonConsoleWidget
    from .console_sci import ShellScintilla


class writeOut(QObject):
    # QsciLexerPython uses style codes up to 15 (Decorator style). We use 16 for error messages
    ERROR_STYLE_INDEX = 16
    ERROR_COLOR = "#e31a1c"

    def __init__(self, shellOut, out=None, style=None):
        """
        This class allows writing to stdout and stderr
        """
        super().__init__()
        self.sO = shellOut
        self.out = None
        self.style = style
        self.fire_keyboard_interrupt = False

    @pyqtSlot(str)
    def write(self, m):
        if sip.isdeleted(self.sO):
            return

        # This manage the case when console is called from another thread
        if QThread.currentThread() != QCoreApplication.instance().thread():
            QMetaObject.invokeMethod(
                self, "write", Qt.ConnectionType.QueuedConnection, Q_ARG(str, m)
            )
            return

        if self.style == "_traceback":
            # Show errors in red
            stderrColor = QColor(
                QgsSettings().value(
                    "pythonConsole/stderrFontColor", QColor(self.ERROR_COLOR)
                )
            )
            self.sO.SendScintilla(
                QsciScintilla.SCI_STYLESETFORE, self.ERROR_STYLE_INDEX, stderrColor
            )
            self.sO.SendScintilla(
                QsciScintilla.SCI_STYLESETITALIC, self.ERROR_STYLE_INDEX, True
            )
            self.sO.SendScintilla(
                QsciScintilla.SCI_STYLESETBOLD, self.ERROR_STYLE_INDEX, True
            )
            pos = self.sO.linearPosition()
            self.sO.SendScintilla(QsciScintilla.SCI_STARTSTYLING, pos, 0)
            self.sO.append(m)
            self.sO.SendScintilla(
                QsciScintilla.SCI_SETSTYLING, len(m), self.ERROR_STYLE_INDEX
            )

        else:
            self.sO.append(m)

        if self.out:
            self.out.write(m)

        self.sO.moveCursorToEnd()

        if self.style != "_traceback":
            self.sO.repaint()

        if self.fire_keyboard_interrupt:
            self.fire_keyboard_interrupt = False
            raise KeyboardInterrupt

    def flush(self):
        pass

    def isatty(self):
        return False


FULL_HELP_TEXT = QCoreApplication.translate(
    "PythonConsole",
    """QGIS Python Console
======================================

The console is a Python interpreter that allows you to execute python commands.
Modules from QGIS (analysis, core, gui, 3d) and Qt (QtCore, QtGui, QtNetwork,
QtWidgets, QtXml) as well as Python's math, os, re and sys modules are already
imported and can be used directly.

Useful variables:

- iface will return the current QGIS interface, class 'QgisInterface'
- iface.mainWindow() will return the Qt Main Window
- iface.mapCanvas() will return the map canvas
- iface.layerTreeView() will return the layer tree
- iface.activeLayer() will return the active layer
- QgsProject.instance() will return the current project

From the console, you can type the following special commands:

    - _pyqgis, _pyqgis(object): Open the QGIS Python API (or the Qt documentation) in a web browser
    - _api, _api(object): Open the QGIS C++ API (or the Qt documentation) in a web browser
    - _cookbook: Open the PyQGIS Developer Cookbook in a web browser
    - System commands: Any command starting with an exclamation mark (!) will be executed by the system shell. Examples:
        !gdalinfo --formats: List all available GDAL drivers
        !ogr2ogr --help: Show help for the ogr2ogr command
        !ping www.qgis.org: Ping the QGIS website
        !pip install black: install black python formatter using pip (if available)
    - ?: Show this help
""",
)


class ShellOutputScintilla(QgsCodeEditorPython):
    def __init__(
        self, console_widget: PythonConsoleWidget, shell_editor: ShellScintilla
    ):
        super().__init__(console_widget)
        self.console_widget: PythonConsoleWidget = console_widget
        self.shell_editor: ShellScintilla = shell_editor

        # Creates layout for message bar
        self.layout = QGridLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        spacerItem = QSpacerItem(
            20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
        )
        self.layout.addItem(spacerItem, 1, 0, 1, 1)
        # messageBar instance
        self.infoBar = QgsMessageBar()
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.infoBar.setSizePolicy(sizePolicy)
        self.layout.addWidget(self.infoBar, 0, 0, 1, 1)

        self._old_stdout = sys.stdout
        self._old_stderr = sys.stderr

        sys.stdout = writeOut(self, sys.stdout)
        sys.stderr = writeOut(self, sys.stderr, "_traceback")

        QgsApplication.instance().aboutToQuit.connect(self.on_app_exit)

        self.insertInitText()
        self.refreshSettingsOutput()

        self.setMinimumHeight(120)

        self.setWrapMode(QsciScintilla.WrapMode.WrapCharacter)
        self.SendScintilla(QsciScintilla.SCI_SETHSCROLLBAR, 0)

        self.runScut = QShortcut(QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_E), self)
        self.runScut.setContext(Qt.ShortcutContext.WidgetShortcut)
        self.runScut.activated.connect(self.enteredSelected)
        # Reimplemented copy action to prevent paste prompt (>>>,...) in command view
        self.copyShortcut = QShortcut(QKeySequence.StandardKey.Copy, self)
        self.copyShortcut.setContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
        self.copyShortcut.activated.connect(self.copy)
        self.selectAllShortcut = QShortcut(QKeySequence.StandardKey.SelectAll, self)
        self.selectAllShortcut.setContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
        self.selectAllShortcut.activated.connect(self.selectAll)

    def on_app_exit(self):
        """
        Prepares the console for a graceful close
        """
        sys.stdout = self._old_stdout
        sys.stderr = self._old_stderr

    def insertInitText(self):
        txtInit = QCoreApplication.translate(
            "PythonConsole",
            "Python Console\n"
            "Use iface to access QGIS API interface or type '?' for more info\n"
            "Security warning: typing commands from an untrusted source can harm your computer",
        )

        txtInit = "\n".join(["# " + line for line in txtInit.split("\n")])

        # some translation string for the console header ends without '\n'
        # and the first command in console will be appended at the header text.
        # The following code add a '\n' at the end of the string if not present.
        if txtInit.endswith("\n"):
            self.setText(txtInit)
        else:
            self.setText(txtInit + "\n")

    def insertHelp(self):
        self.append(FULL_HELP_TEXT)
        self.moveCursorToEnd()

    def initializeLexer(self):
        super().initializeLexer()
        self.setFoldingVisible(False)
        self.setEdgeMode(QsciScintilla.EdgeMode.EdgeNone)

    def refreshSettingsOutput(self):
        # Set Python lexer
        self.initializeLexer()
        self.setReadOnly(True)

        self.setCaretWidth(0)  # NO (blinking) caret in the output

    def clearConsole(self):
        self.setText("")
        self.insertInitText()
        self.shell_editor.setFocus()

    def contextMenuEvent(self, e):
        menu = QMenu(self)
        menu.addAction(
            QgsApplication.getThemeIcon("console/iconHideToolConsole.svg"),
            QCoreApplication.translate("PythonConsole", "Hide/Show Toolbar"),
            self.hideToolBar,
        )
        menu.addSeparator()
        showEditorAction = menu.addAction(
            QgsApplication.getThemeIcon("console/iconShowEditorConsole.svg"),
            QCoreApplication.translate("PythonConsole", "Show Editor"),
            self.showEditor,
        )
        menu.addSeparator()
        runAction = QAction(
            QgsApplication.getThemeIcon("console/mIconRunConsole.svg"),
            QCoreApplication.translate("PythonConsole", "Enter Selected"),
            menu,
        )
        runAction.triggered.connect(self.enteredSelected)
        runAction.setShortcut(QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_E))
        menu.addAction(runAction)

        clearAction = QAction(
            QgsApplication.getThemeIcon("console/iconClearConsole.svg"),
            QCoreApplication.translate("PythonConsole", "Clear Console"),
            menu,
        )
        clearAction.triggered.connect(self.clearConsole)
        menu.addAction(clearAction)

        word = self.selectedText() or self.wordAtPoint(e.pos())
        if word:
            context_help_action = QAction(
                QgsApplication.getThemeIcon("mActionHelpContents.svg"),
                QCoreApplication.translate("PythonConsole", "Context Help"),
                menu,
            )
            context_help_action.triggered.connect(
                partial(self.shell_editor.showApiDocumentation, word, force_search=True)
            )
            context_help_action.setShortcut(QKeySequence.StandardKey.HelpContents)
            menu.addAction(context_help_action)

        menu.addSeparator()
        copyAction = QAction(
            QgsApplication.getThemeIcon("mActionEditCopy.svg"),
            QCoreApplication.translate("PythonConsole", "Copy"),
            menu,
        )
        copyAction.triggered.connect(self.copy)
        copyAction.setShortcut(QKeySequence.StandardKey.Copy)
        menu.addAction(copyAction)

        selectAllAction = QAction(
            QCoreApplication.translate("PythonConsole", "Select All"), menu
        )
        selectAllAction.triggered.connect(self.selectAll)
        selectAllAction.setShortcut(QKeySequence.StandardKey.SelectAll)
        menu.addAction(selectAllAction)

        menu.addSeparator()
        settings_action = QAction(
            QgsApplication.getThemeIcon("console/iconSettingsConsole.svg"),
            QCoreApplication.translate("PythonConsole", "Options…"),
            menu,
        )
        settings_action.triggered.connect(self.console_widget.openSettings)
        menu.addAction(settings_action)

        runAction.setEnabled(False)
        clearAction.setEnabled(False)
        copyAction.setEnabled(False)
        selectAllAction.setEnabled(False)
        showEditorAction.setEnabled(True)
        if self.hasSelectedText():
            runAction.setEnabled(True)
            copyAction.setEnabled(True)
        if not self.text(3) == "":
            selectAllAction.setEnabled(True)
            clearAction.setEnabled(True)
        if self.console_widget.tabEditorWidget.isVisible():
            showEditorAction.setEnabled(False)
        menu.exec(self.mapToGlobal(e.pos()))

    def hideToolBar(self):
        tB = self.console_widget.toolBar
        tB.hide() if tB.isVisible() else tB.show()
        self.shell_editor.setFocus()

    def showEditor(self):
        Ed = self.console_widget.splitterObj
        if not Ed.isVisible():
            Ed.show()
            self.console_widget.show_editor_action.setChecked(True)
        self.shell_editor.setFocus()

    def copy(self):
        """Copy text to clipboard... or keyboard interrupt"""
        if self.hasSelectedText():
            text = self.selectedText()
            text = (
                text.replace(">>> ", "").replace("... ", "").strip()
            )  # removing prompts
            QApplication.clipboard().setText(text)
        else:
            raise KeyboardInterrupt

    def enteredSelected(self):
        cmd = self.selectedText()
        self.shell_editor.insertFromDropPaste(cmd)
        self.shell_editor.entered()

    def widgetMessageBar(self, text: str):
        self.infoBar.pushMessage(text, Qgis.MessageLevel.Info)

    def showApiDocumentation(self, text):
        self.shell_editor.showApiDocumentation(text)



"""
***************************************************************************
    ogrinfo.py
    ---------------------
    Date                 : November 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "November 2012"
__copyright__ = "(C) 2012, Victor Olaya"

from qgis.core import (
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterString,
    QgsProcessingParameterVectorLayer,
)

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils


class ogrinfo(GdalAlgorithm):
    INPUT = "INPUT"
    ALL_LAYERS = "ALL_LAYERS"
    SUMMARY_ONLY = "SUMMARY_ONLY"
    NO_METADATA = "NO_METADATA"
    EXTRA = "EXTRA"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterVectorLayer(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.ALL_LAYERS,
                self.tr("Enable listing of all layers in the dataset"),
                defaultValue=False,
            )
        )
        if self.name() == "ogrinfo":
            self.addParameter(
                QgsProcessingParameterBoolean(
                    self.SUMMARY_ONLY, self.tr("Summary output only"), defaultValue=True
                )
            )
        else:
            self.addParameter(
                QgsProcessingParameterBoolean(
                    self.FEATURES,
                    self.tr("Enable listing of features"),
                    defaultValue=False,
                )
            )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.NO_METADATA, self.tr("Suppress metadata info"), defaultValue=False
            )
        )

        extra_param = QgsProcessingParameterString(
            self.EXTRA,
            self.tr("Additional command-line parameters"),
            defaultValue=None,
            optional=True,
        )
        extra_param.setFlags(
            extra_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(extra_param)

        if self.name() == "ogrinfo":
            self.addParameter(
                QgsProcessingParameterFileDestination(
                    self.OUTPUT,
                    self.tr("Layer information"),
                    self.tr("HTML files (*.html)"),
                )
            )
        else:
            self.addParameter(
                QgsProcessingParameterFileDestination(
                    self.OUTPUT,
                    self.tr("Layer information"),
                    self.tr("JSON files (*.json)"),
                )
            )

    def name(self):
        return "ogrinfo"

    def displayName(self):
        return self.tr("Vector information")

    def group(self):
        return self.tr("Vector miscellaneous")

    def groupId(self):
        return "vectormiscellaneous"

    def commandName(self):
        return "ogrinfo"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        if self.name() == "ogrinfo":
            arguments = ["-al"]
        else:
            arguments = ["-json"]

        if self.name() == "ogrinfo":
            if self.parameterAsBoolean(parameters, self.SUMMARY_ONLY, context):
                arguments.append("-so")
        else:
            if self.parameterAsBoolean(parameters, self.FEATURES, context):
                arguments.append("-features")

        if self.parameterAsBoolean(parameters, self.NO_METADATA, context):
            arguments.append("-nomd")

        if self.EXTRA in parameters and parameters[self.EXTRA] not in (None, ""):
            extra = self.parameterAsString(parameters, self.EXTRA, context)
            arguments.append(extra)

        inLayer = self.parameterAsVectorLayer(parameters, self.INPUT, context)
        if inLayer is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        input_details = self.getOgrCompatibleSource(
            self.INPUT, parameters, context, feedback, executing
        )
        arguments.append(input_details.connection_string)
        if not self.parameterAsBoolean(parameters, self.ALL_LAYERS, context):
            arguments.append(input_details.layer_name)

        if input_details.open_options:
            arguments.extend(input_details.open_options_as_arguments())

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]

    def processAlgorithm(self, parameters, context, feedback):
        console_output = GdalUtils.runGdal(
            self.getConsoleCommands(parameters, context, feedback), feedback
        )
        output = self.parameterAsFileOutput(parameters, self.OUTPUT, context)
        with open(output, "w") as f:
            f.write("<pre>")
            for s in console_output[1:]:
                f.write(str(s))
            f.write("</pre>")

        return {self.OUTPUT: output}


class ogrinfojson(ogrinfo):
    FEATURES = "FEATURES"

    def name(self):
        return "ogrinfojson"

    def displayName(self):
        return self.tr("Vector information (JSON)")

    def processAlgorithm(self, parameters, context, feedback):
        console_output = GdalUtils.runGdal(
            self.getConsoleCommands(parameters, context, feedback)
        )
        output = self.parameterAsFileOutput(parameters, self.OUTPUT, context)
        with open(output, "w", newline="") as f:
            for s in console_output[1:]:
                f.write(str(s))

        return {self.OUTPUT: output}



"""
***************************************************************************
    OgrToPostGis.py
    ---------------------
    Date                 : November 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "November 2012"
__copyright__ = "(C) 2012, Victor Olaya"

from qgis.core import (
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterCrs,
    QgsProcessingParameterEnum,
    QgsProcessingParameterExtent,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterString,
)

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils
from processing.tools.system import isWindows


class OgrToPostGis(GdalAlgorithm):
    INPUT = "INPUT"
    SHAPE_ENCODING = "SHAPE_ENCODING"
    GTYPE = "GTYPE"
    GEOMTYPE = [
        "",
        "NONE",
        "GEOMETRY",
        "POINT",
        "LINESTRING",
        "POLYGON",
        "GEOMETRYCOLLECTION",
        "MULTIPOINT",
        "MULTIPOLYGON",
        "MULTILINESTRING",
        "CIRCULARSTRING",
        "COMPOUNDCURVE",
        "CURVEPOLYGON",
        "MULTICURVE",
        "MULTISURFACE",
        "CONVERT_TO_LINEAR",
        "CONVERT_TO_CURVE",
    ]
    S_SRS = "S_SRS"
    T_SRS = "T_SRS"
    A_SRS = "A_SRS"
    HOST = "HOST"
    PORT = "PORT"
    USER = "USER"
    DBNAME = "DBNAME"
    PASSWORD = "PASSWORD"
    SCHEMA = "SCHEMA"
    TABLE = "TABLE"
    PK = "PK"
    PRIMARY_KEY = "PRIMARY_KEY"
    GEOCOLUMN = "GEOCOLUMN"
    DIM = "DIM"
    DIMLIST = ["2", "3", "4"]
    SIMPLIFY = "SIMPLIFY"
    SEGMENTIZE = "SEGMENTIZE"
    SPAT = "SPAT"
    CLIP = "CLIP"
    FIELDS = "FIELDS"
    WHERE = "WHERE"
    GT = "GT"
    OVERWRITE = "OVERWRITE"
    APPEND = "APPEND"
    ADDFIELDS = "ADDFIELDS"
    LAUNDER = "LAUNDER"
    INDEX = "INDEX"
    SKIPFAILURES = "SKIPFAILURES"
    PRECISION = "PRECISION"
    MAKEVALID = "MAKEVALID"
    PROMOTETOMULTI = "PROMOTETOMULTI"
    OPTIONS = "OPTIONS"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT,
                self.tr("Input layer"),
                types=[QgsProcessing.SourceType.TypeVector],
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.SHAPE_ENCODING, self.tr("Shape encoding"), "", optional=True
            )
        )
        self.addParameter(
            QgsProcessingParameterEnum(
                self.GTYPE,
                self.tr("Output geometry type"),
                options=self.GEOMTYPE,
                defaultValue=0,
            )
        )
        self.addParameter(
            QgsProcessingParameterCrs(
                self.A_SRS,
                self.tr("Assign an output CRS"),
                defaultValue="",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterCrs(
                self.T_SRS,
                self.tr("Reproject to this CRS on output "),
                defaultValue="",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterCrs(
                self.S_SRS,
                self.tr("Override source CRS"),
                defaultValue="",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.HOST, self.tr("Host"), defaultValue="localhost", optional=True
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.PORT, self.tr("Port"), defaultValue="5432", optional=True
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.USER, self.tr("Username"), defaultValue="", optional=True
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.DBNAME, self.tr("Database name"), defaultValue="", optional=True
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.PASSWORD, self.tr("Password"), defaultValue="", optional=True
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.SCHEMA,
                self.tr("Schema name"),
                defaultValue="public",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.TABLE,
                self.tr("Table name, leave blank to use input name"),
                defaultValue="",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.PK,
                self.tr("Primary key (new field)"),
                defaultValue="id",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.PRIMARY_KEY,
                self.tr(
                    "Primary key (existing field, used if the above option is left empty)"
                ),
                parentLayerParameterName=self.INPUT,
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.GEOCOLUMN,
                self.tr("Geometry column name"),
                defaultValue="geom",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterEnum(
                self.DIM,
                self.tr("Vector dimensions"),
                options=self.DIMLIST,
                defaultValue=0,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.SIMPLIFY,
                self.tr("Distance tolerance for simplification"),
                defaultValue="",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.SEGMENTIZE,
                self.tr("Maximum distance between 2 nodes (densification)"),
                defaultValue="",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterExtent(
                self.SPAT,
                self.tr("Select features by extent (defined in input layer CRS)"),
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.CLIP,
                self.tr("Clip the input layer using the above (rectangle) extent"),
                defaultValue=False,
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.FIELDS,
                self.tr("Fields to include (leave empty to use all fields)"),
                parentLayerParameterName=self.INPUT,
                allowMultiple=True,
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.WHERE,
                self.tr(
                    "Select features using a SQL \"WHERE\" statement (Ex: column='value')"
                ),
                defaultValue="",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.GT,
                self.tr("Group N features per transaction (Default: 20000)"),
                defaultValue="",
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.OVERWRITE, self.tr("Overwrite existing table"), defaultValue=True
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.APPEND, self.tr("Append to existing table"), defaultValue=False
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.ADDFIELDS,
                self.tr("Append and add new fields to existing table"),
                defaultValue=False,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.LAUNDER,
                self.tr("Do not launder columns/table names"),
                defaultValue=False,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.INDEX, self.tr("Do not create spatial index"), defaultValue=False
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.SKIPFAILURES,
                self.tr("Continue after a failure, skipping the failed feature"),
                defaultValue=False,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.MAKEVALID,
                self.tr("Validate geometries based on Simple Features specification"),
                defaultValue=False,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.PROMOTETOMULTI, self.tr("Promote to Multipart"), defaultValue=True
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.PRECISION,
                self.tr("Keep width and precision of input attributes"),
                defaultValue=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.OPTIONS,
                self.tr("Additional creation options"),
                defaultValue="",
                optional=True,
            )
        )

    def name(self):
        return "importvectorintopostgisdatabasenewconnection"

    def displayName(self):
        return self.tr("Export to PostgreSQL (new connection)")

    def shortDescription(self):
        return self.tr("Exports a vector layer to a new PostgreSQL database connection")

    def tags(self):
        t = self.tr("import,into,postgis,database,vector").split(",")
        t.extend(super().tags())
        return t

    def group(self):
        return self.tr("Vector miscellaneous")

    def groupId(self):
        return "vectormiscellaneous"

    def getConnectionString(self, parameters, context):
        host = self.parameterAsString(parameters, self.HOST, context)
        port = self.parameterAsString(parameters, self.PORT, context)
        user = self.parameterAsString(parameters, self.USER, context)
        dbname = self.parameterAsString(parameters, self.DBNAME, context)
        password = self.parameterAsString(parameters, self.PASSWORD, context)
        schema = self.parameterAsString(parameters, self.SCHEMA, context)
        arguments = []
        if host:
            arguments.append("host=" + host)
        if port:
            arguments.append("port=" + str(port))
        if dbname:
            arguments.append("dbname=" + dbname)
        if password:
            arguments.append("password=" + password)
        if schema:
            arguments.append("active_schema=" + schema)
        if user:
            arguments.append("user=" + user)
        return GdalUtils.escapeAndJoin(arguments)

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        input_details = self.getOgrCompatibleSource(
            self.INPUT, parameters, context, feedback, executing
        )
        if not input_details.layer_name:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        shapeEncoding = self.parameterAsString(parameters, self.SHAPE_ENCODING, context)
        ssrs = self.parameterAsCrs(parameters, self.S_SRS, context)
        tsrs = self.parameterAsCrs(parameters, self.T_SRS, context)
        asrs = self.parameterAsCrs(parameters, self.A_SRS, context)
        table = self.parameterAsString(parameters, self.TABLE, context)
        schema = self.parameterAsString(parameters, self.SCHEMA, context)
        pk = self.parameterAsString(parameters, self.PK, context)
        pkstring = "-lco FID=" + pk
        primary_key = self.parameterAsString(parameters, self.PRIMARY_KEY, context)

        geocolumn = self.parameterAsString(parameters, self.GEOCOLUMN, context)
        if input_details.geometry_column_name:
            geocolumn = input_details.geometry_column_name

        geocolumnstring = "-lco GEOMETRY_NAME=" + geocolumn
        dim = self.DIMLIST[self.parameterAsEnum(parameters, self.DIM, context)]
        dimstring = "-lco DIM=" + dim
        simplify = self.parameterAsString(parameters, self.SIMPLIFY, context)
        segmentize = self.parameterAsString(parameters, self.SEGMENTIZE, context)
        spat = self.parameterAsExtent(parameters, self.SPAT, context)
        clip = self.parameterAsBoolean(parameters, self.CLIP, context)
        include_fields = self.parameterAsFields(parameters, self.FIELDS, context)
        fields_string = '-select "' + ",".join(include_fields) + '"'
        where = self.parameterAsString(parameters, self.WHERE, context)
        wherestring = '-where "' + where + '"'
        gt = self.parameterAsString(parameters, self.GT, context)
        overwrite = self.parameterAsBoolean(parameters, self.OVERWRITE, context)
        append = self.parameterAsBoolean(parameters, self.APPEND, context)
        addfields = self.parameterAsBoolean(parameters, self.ADDFIELDS, context)
        launder = self.parameterAsBoolean(parameters, self.LAUNDER, context)
        launderstring = "-lco LAUNDER=NO"
        index = self.parameterAsBoolean(parameters, self.INDEX, context)
        indexstring = "-lco SPATIAL_INDEX=OFF"
        skipfailures = self.parameterAsBoolean(parameters, self.SKIPFAILURES, context)
        make_valid = self.parameterAsBoolean(parameters, self.MAKEVALID, context)
        promotetomulti = self.parameterAsBoolean(
            parameters, self.PROMOTETOMULTI, context
        )
        precision = self.parameterAsBoolean(parameters, self.PRECISION, context)
        options = self.parameterAsString(parameters, self.OPTIONS, context)

        arguments = ["-progress", "--config PG_USE_COPY YES"]
        if len(shapeEncoding) > 0:
            arguments.append("--config")
            arguments.append("SHAPE_ENCODING")
            arguments.append(shapeEncoding)
        arguments.append("-f")
        arguments.append("PostgreSQL")
        arguments.append("PG:" + self.getConnectionString(parameters, context))
        arguments.append(dimstring)
        arguments.append(input_details.connection_string)
        arguments.append(input_details.layer_name)
        if index:
            arguments.append(indexstring)
        if launder:
            arguments.append(launderstring)
        if append and overwrite:
            raise QgsProcessingException(
                self.tr(
                    'Only one of "Overwrite existing table" or "Append to existing table" can be enabled at a time.'
                )
            )
        elif append:
            arguments.append("-append")
        if include_fields:
            arguments.append(fields_string)
        if addfields:
            arguments.append("-addfields")
        if overwrite:
            arguments.append("-overwrite")
        if (
            len(self.GEOMTYPE[self.parameterAsEnum(parameters, self.GTYPE, context)])
            > 0
        ):
            arguments.append("-nlt")
            arguments.append(
                self.GEOMTYPE[self.parameterAsEnum(parameters, self.GTYPE, context)]
            )
        if len(geocolumn) > 0:
            arguments.append(geocolumnstring)
        if pk:
            arguments.append(pkstring)
        elif primary_key:
            arguments.append("-lco FID=" + primary_key)
        if len(table) == 0:
            table = input_details.layer_name.lower()
        if schema:
            table = f"{schema}.{table}"
        arguments.append("-nln")
        arguments.append(table)
        if ssrs.isValid():
            arguments.append("-s_srs")
            arguments.append(GdalUtils.gdal_crs_string(ssrs))
        if tsrs.isValid():
            arguments.append("-t_srs")
            arguments.append(GdalUtils.gdal_crs_string(tsrs))
        if asrs.isValid():
            arguments.append("-a_srs")
            arguments.append(GdalUtils.gdal_crs_string(asrs))
        if not spat.isNull():
            arguments.append("-spat")
            arguments.append(spat.xMinimum())
            arguments.append(spat.yMinimum())
            arguments.append(spat.xMaximum())
            arguments.append(spat.yMaximum())
            if clip:
                arguments.append("-clipsrc spat_extent")
        if skipfailures:
            arguments.append("-skipfailures")
        if where:
            arguments.append(wherestring)
        if len(simplify) > 0:
            arguments.append("-simplify")
            arguments.append(simplify)
        if len(segmentize) > 0:
            arguments.append("-segmentize")
            arguments.append(segmentize)
        if len(gt) > 0:
            arguments.append("-gt")
            arguments.append(gt)
        if make_valid:
            arguments.append("-makevalid")
        if (
            promotetomulti
            and self.GEOMTYPE[self.parameterAsEnum(parameters, self.GTYPE, context)]
        ):
            if (
                self.GEOMTYPE[self.parameterAsEnum(parameters, self.GTYPE, context)]
                == "CONVERT_TO_LINEAR"
            ):
                arguments.append("-nlt PROMOTE_TO_MULTI")
            else:
                raise QgsProcessingException(
                    self.tr(
                        'Only one of "Promote to Multipart" or "Output geometry type" (excluding Convert to Linear) can be enabled.'
                    )
                )
        elif (
            promotetomulti
            and not self.GEOMTYPE[self.parameterAsEnum(parameters, self.GTYPE, context)]
        ):
            arguments.append("-nlt PROMOTE_TO_MULTI")
        if precision is False:
            arguments.append("-lco PRECISION=NO")

        if input_details.open_options:
            arguments.extend(input_details.open_options_as_arguments())

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        if len(options) > 0:
            arguments.append(options)

        if isWindows():
            return ["cmd.exe", "/C ", "ogr2ogr.exe", GdalUtils.escapeAndJoin(arguments)]
        else:
            return ["ogr2ogr", GdalUtils.escapeAndJoin(arguments)]

    def commandName(self):
        return "ogr2ogr"



"""Parse plugin metadata for plugin_dependencies

.. note:: This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

"""

__author__ = "elpaso@itopen.it"
__date__ = "2018-05-29"
__copyright__ = "Copyright 2018, GISCE-TI S.L."

from configparser import NoOptionError, NoSectionError

from qgis.utils import get_plugin_deps, metadataParser, updateAvailablePlugins

from . import installer as plugin_installer
from .version_compare import compareVersions


def __plugin_name_map(plugin_data_values):
    return {plugin["name"]: plugin["id"] for plugin in plugin_data_values}


def find_dependencies(
    plugin_id, plugin_data=None, plugin_deps=None, installed_plugins=None
):
    """Finds the plugin dependencies and checks if they can be installed or upgraded

    :param plugin_id: plugin id
    :type plugin_id: str
    :param plugin_data: for testing only: dictionary of plugin data from the repo, defaults to None
    :param plugin_data: dict, optional
    :param plugin_deps: for testing only: dict of plugin id -> version_required, parsed from metadata value for "plugin_dependencies", defaults to None
    :param plugin_deps: dict, optional
    :param installed_plugins: for testing only: dict of plugin id -> version_installed
    :param installed_plugins: dict, optional
    :return: result dictionaries keyed by plugin name with: to_install, to_upgrade, not_found
    :rtype: tuple of dicts
    """

    to_install = {}
    to_upgrade = {}
    not_found = {}

    if plugin_deps is None or installed_plugins is None:
        updateAvailablePlugins()

    if plugin_deps is None:
        plugin_deps = get_plugin_deps(plugin_id)

    if installed_plugins is None:
        metadata_parser = metadataParser()
        installed_plugins = {
            metadata_parser[k].get("general", "name"): metadata_parser[k].get(
                "general", "version"
            )
            for k, v in metadata_parser.items()
        }

    if plugin_data is None:
        plugin_data = plugin_installer.plugins.all()

    plugins_map = __plugin_name_map(plugin_data.values())

    # Review all dependencies
    for name, version_required in plugin_deps.items():
        try:
            p_id = plugins_map[name]
        except KeyError:
            not_found.update(
                {
                    name: {
                        "id": None,
                        "version_installed": None,
                        "version_required": None,
                        "version_available": None,
                        "use_stable_version": None,
                        "action": None,
                        "error": "missing_id",
                    }
                }
            )
            continue

        affected_plugin = dict(
            {
                "id": p_id,
                # "version_installed": installed_plugins.get(p_id, {}).get('installed_plugins', None),
                "version_installed": installed_plugins.get(name, None),
                "version_required": version_required,
                "version_available": plugin_data[p_id].get("version_available", None),
                "use_stable_version": True,  # Prefer stable by default
                "action": None,
            }
        )
        version_available_stable = plugin_data[p_id].get(
            "version_available_stable", None
        )
        version_available_experimental = plugin_data[p_id].get(
            "version_available_experimental", None
        )

        if (
            version_required is not None
            and version_required == version_available_stable
        ):
            affected_plugin["version_available"] = version_available_stable
            affected_plugin["use_stable_version"] = True
        elif (
            version_required is not None
            and version_required == version_available_experimental
        ):
            affected_plugin["version_available"] = version_available_experimental
            affected_plugin["use_stable_version"] = False
        elif version_required is None:
            if version_available_stable:  # None if not found, "" if not offered
                # Prefer the stable version, if any
                affected_plugin["version_available"] = version_available_stable
                affected_plugin["use_stable_version"] = True
            else:  # The only available version is experimental
                affected_plugin["version_available"] = version_available_experimental
                affected_plugin["use_stable_version"] = False

        # Install is needed
        if name not in installed_plugins:
            affected_plugin["action"] = "install"
            destination_list = to_install
        # Upgrade is needed
        elif (
            version_required is not None
            and compareVersions(installed_plugins[name], version_required) == 2
        ):
            affected_plugin["action"] = "upgrade"
            destination_list = to_upgrade
        # TODO @elpaso: review installed but not activated
        # No action is needed
        else:
            continue

        if (
            version_required == affected_plugin["version_available"]
            or version_required is None
        ):
            destination_list.update({name: affected_plugin})
        else:
            affected_plugin["error"] = "unavailable {}".format(
                affected_plugin["action"]
            )
            not_found.update({name: affected_plugin})

    return to_install, to_upgrade, not_found



"""
***************************************************************************
    HubDistanceLines.py
    ---------------------
    Date                 : May 2010
    Copyright            : (C) 2010 by Michael Minn
    Email                : pyqgis at michaelminn dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Michael Minn"
__date__ = "May 2010"
__copyright__ = "(C) 2010, Michael Minn"

from math import sqrt

from qgis.core import (
    Qgis,
    QgsDistanceArea,
    QgsFeature,
    QgsFeatureRequest,
    QgsFeatureSink,
    QgsField,
    QgsFields,
    QgsGeometry,
    QgsProcessing,
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingUtils,
    QgsSpatialIndex,
    QgsUnitTypes,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QMetaType

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm


class HubDistanceLines(QgisAlgorithm):
    INPUT = "INPUT"
    HUBS = "HUBS"
    FIELD = "FIELD"
    UNIT = "UNIT"
    OUTPUT = "OUTPUT"

    LAYER_UNITS = "LAYER_UNITS"

    UNITS = [
        QgsUnitTypes.DistanceUnit.DistanceMeters,
        QgsUnitTypes.DistanceUnit.DistanceFeet,
        QgsUnitTypes.DistanceUnit.DistanceMiles,
        QgsUnitTypes.DistanceUnit.DistanceKilometers,
        LAYER_UNITS,
    ]

    def group(self):
        return self.tr("Vector analysis")

    def groupId(self):
        return "vectoranalysis"

    def documentationFlags(self):
        return Qgis.ProcessingAlgorithmDocumentationFlag.RespectsEllipsoid

    def __init__(self):
        super().__init__()

    def flags(self):
        return super().flags() | QgsProcessingAlgorithm.Flag.FlagDeprecated

    def initAlgorithm(self, config=None):
        self.units = [
            self.tr("Meters"),
            self.tr("Feet"),
            self.tr("Miles"),
            self.tr("Kilometers"),
            self.tr("Layer units"),
        ]

        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT, self.tr("Source points layer")
            )
        )
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.HUBS, self.tr("Destination hubs layer")
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.FIELD,
                self.tr("Hub layer name attribute"),
                parentLayerParameterName=self.HUBS,
            )
        )
        self.addParameter(
            QgsProcessingParameterEnum(
                self.UNIT, self.tr("Measurement unit"), self.units
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr("Hub distance"),
                QgsProcessing.SourceType.TypeVectorLine,
            )
        )

    def name(self):
        return "distancetonearesthublinetohub"

    def displayName(self):
        return self.tr("Distance to nearest hub (line to hub)")

    def shortDescription(self):
        return self.tr(
            "Creates lines that join each feature from an input vector to the nearest feature in a destination layer."
        )

    def shortHelpString(self):
        return self.tr(
            "Given an origin and a destination layers, this algorithm computes "
            "the distance between origin features and their closest destination one. "
            "Distance calculations are based on the feature's center.\n"
            "The resulting layer contains lines linking each origin point with its nearest destination feature."
        )

    def processAlgorithm(self, parameters, context, feedback):
        if parameters[self.INPUT] == parameters[self.HUBS]:
            raise QgsProcessingException(
                self.tr("Same layer given for both hubs and spokes")
            )

        point_source = self.parameterAsSource(parameters, self.INPUT, context)
        if point_source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        hub_source = self.parameterAsSource(parameters, self.HUBS, context)
        if hub_source is None:
            raise QgsProcessingException(self.invalidSourceError(parameters, self.HUBS))

        fieldName = self.parameterAsString(parameters, self.FIELD, context)

        units = self.UNITS[self.parameterAsEnum(parameters, self.UNIT, context)]

        newFields = QgsFields()
        newFields.append(QgsField("HubName", QMetaType.Type.QString))
        newFields.append(QgsField("HubDist", QMetaType.Type.Double))
        fields = QgsProcessingUtils.combineFields(point_source.fields(), newFields)

        (sink, dest_id) = self.parameterAsSink(
            parameters,
            self.OUTPUT,
            context,
            fields,
            QgsWkbTypes.Type.LineString,
            point_source.sourceCrs(),
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        index = QgsSpatialIndex(
            hub_source.getFeatures(
                QgsFeatureRequest()
                .setSubsetOfAttributes([])
                .setDestinationCrs(point_source.sourceCrs(), context.transformContext())
            )
        )

        distance = QgsDistanceArea()
        distance.setSourceCrs(point_source.sourceCrs(), context.transformContext())
        distance.setEllipsoid(context.ellipsoid())

        # Scan source points, find nearest hub, and write to output file
        features = point_source.getFeatures()
        total = (
            100.0 / point_source.featureCount() if point_source.featureCount() else 0
        )
        for current, f in enumerate(features):
            if feedback.isCanceled():
                break

            if not f.hasGeometry():
                sink.addFeature(f, QgsFeatureSink.Flag.FastInsert)
                continue
            src = f.geometry().boundingBox().center()

            neighbors = index.nearestNeighbor(src, 1)
            if len(neighbors) == 0:
                continue

            ft = next(
                hub_source.getFeatures(
                    QgsFeatureRequest()
                    .setFilterFid(neighbors[0])
                    .setSubsetOfAttributes([fieldName], hub_source.fields())
                    .setDestinationCrs(
                        point_source.sourceCrs(), context.transformContext()
                    )
                )
            )
            closest = ft.geometry().boundingBox().center()
            hubDist = distance.measureLine(src, closest)

            if units != self.LAYER_UNITS:
                hub_dist_in_desired_units = distance.convertLengthMeasurement(
                    hubDist, units
                )
            else:
                hub_dist_in_desired_units = hubDist

            attributes = f.attributes()
            attributes.append(ft[fieldName])
            attributes.append(hub_dist_in_desired_units)

            feat = QgsFeature()
            feat.setAttributes(attributes)

            feat.setGeometry(QgsGeometry.fromPolylineXY([src, closest]))

            sink.addFeature(feat, QgsFeatureSink.Flag.FastInsert)
            feedback.setProgress(int(current * total))

        sink.finalize()
        return {self.OUTPUT: dest_id}
